import datetime
from peewee import *
from troi.content_resolver.model.database import db


class UnresolvedRecording(Model):
    """
    Table used to track which recordings where resolving failed. This can be used both
    for debugging purposes and to provide the user with a list of 'if you had this
    album, you'd resolve more music' kind of report.
    """

    class Meta:
        database = db
        table_name = "unresolved_recording"

    id = AutoField()
    # Not using the UUIDField here, since it annoyingly removes '-' from the UUID.
    recording_mbid = TextField(null=True, index=True, unique=True)
    lookup_count = IntegerField(null=False, default=1)
    last_updated = DateTimeField(null=False, default=datetime.datetime.now)

    def __repr__(self):
        return "<UnresolvedRecording(%s,%d')>" % (self.recording_mbid, self.count)



from troi.content_resolver.formats.tag_utils import extract_track_number


class TestTagUtils:

    def test_int(self):
        assert extract_track_number(100) == 100
        assert extract_track_number(-100) == -100

    def test_string(self):
        assert extract_track_number("100") == 100

    def test_string_with_slash(self):
        assert extract_track_number("9/12") == 9

    def test_invalid_values(self):
        assert extract_track_number(None) == None
        assert extract_track_number("") == None
        assert extract_track_number("NotANumber") == None
        assert extract_track_number("3.0") == None



from troi import Element, Recording, Playlist, PipelineError
import troi.listenbrainz.recs
import troi.playlist
import troi.filters
import troi.sorts
import troi.musicbrainz.recording_lookup
import troi.musicbrainz.mbid_mapping
from troi.patch import Patch


class RecsPlaylistMakerElement(Element):
    '''
        This element takes in Recordings and spits out a Playlist, which generating a custom name and desc
        for the playlist from the recording data
    '''

    def __init__(self, name, desc, patch_slug=None, user_name=None, max_num_recordings=None, type=None):
        super().__init__()
        self.name = name
        self.desc = desc
        self.patch_slug = patch_slug
        self.user_name = user_name
        self.max_num_recordings = max_num_recordings
        self.type = type

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Playlist]

    def read(self, inputs):
        try:
            model_id = inputs[0][0].listenbrainz["model_id"]
            model_url = inputs[0][0].listenbrainz["model_url"]
            self.name = "Top 100 %s recommendations from model %s" % (self.type, model_id)
            self.desc = """Top 100 %s recommendations generated by model <a href="%s">%s</a>""" % (self.type, model_url, model_id)
        except IndexError:
            pass

        if self.max_num_recordings is not None:
            return [Playlist(name=self.name,
                    description=self.desc,
                    recordings=inputs[0][:self.max_num_recordings],
                    patch_slug=self.patch_slug,
                    user_name=self.user_name)]
        else:
            return [Playlist(name=self.name,
                    description=self.desc,
                    recordings=inputs[0],
                    patch_slug=self.patch_slug,
                    user_name=self.user_name)]


class RecommendationsToPlaylistPatch(Patch):
    """
        See below for description
    """

    def __init__(self, args):
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Save the current recommended tracks for a given user and type (top or similar).

        \b
        USER_NAME: is a MusicBrainz user name that has an account on ListenBrainz.
        TYPE: is The type of daily jam. Must be 'top' or 'similar' or 'raw'.
        """
        return [
            {"type": "argument", "args": ["user_name"]},
            {"type": "argument", "args": ["type"]}
        ]

    @staticmethod
    def outputs():
        return [Recording]

    @staticmethod
    def slug():
        return "recs-to-playlist"

    @staticmethod
    def description():
        return "Save the current recommended tracks for a given user and type (top, similar or raw)."

    def create(self, inputs):
        user_name = inputs['user_name']
        type = inputs['type']

        if type not in ("top", "similar", "raw"):
            raise PipelineError("type must be either 'top' or 'similar' or 'raw'")

        recs = troi.listenbrainz.recs.UserRecordingRecommendationsElement(user_name=user_name,
                                                                          artist_type=type,
                                                                          count=100)
        if type == "top":
            playlist_type = "top artists"
        elif type == "similar":
            playlist_type = "similar artists"
        else:
            playlist_type = "raw"
        pl_maker = RecsPlaylistMakerElement(name="[unknown]",
                                            desc="[unknown]",
                                            patch_slug="saved-recs",
                                            user_name=user_name,
                                            type=playlist_type)
        pl_maker.set_sources(recs)

        return pl_maker



import logging
from collections import defaultdict
import json
from time import sleep


import requests
import spotipy
from troi.tools.utils import AppleMusicAPI

from troi import Recording, Playlist, PipelineError, Element, Artist, ArtistCredit, Release
from troi.operations import is_homogeneous
from troi.print_recording import PrintRecordingList
from troi.tools.common_lookup import music_service_tracks_to_mbid
from troi.tools.spotify_lookup import submit_to_spotify
from troi.tools.apple_lookup import submit_to_apple_music
from troi.tools.soundcloud_lookup import submit_to_soundcloud
from troi.tools.utils import SoundcloudAPI

logger = logging.getLogger(__name__)

LISTENBRAINZ_SERVER_URL = "https://listenbrainz.org"
LISTENBRAINZ_API_URL = "https://api.listenbrainz.org"
LISTENBRAINZ_PLAYLIST_FETCH_URL = LISTENBRAINZ_API_URL + "/1/playlist/"
LISTENBRAINZ_PLAYLIST_CREATE_URL = LISTENBRAINZ_API_URL + "/1/playlist/create"
PLAYLIST_TRACK_URI_PREFIX = "https://musicbrainz.org/recording/"
PLAYLIST_ARTIST_URI_PREFIX = "https://musicbrainz.org/artist/"
PLAYLIST_RELEASE_URI_PREFIX = "https://musicbrainz.org/release/"
PLAYLIST_URI_PREFIX = "https://listenbrainz.org/playlist/"
PLAYLIST_EXTENSION_URI = "https://musicbrainz.org/doc/jspf#playlist"
PLAYLIST_TRACK_EXTENSION_URI = "https://musicbrainz.org/doc/jspf#track"
SUBSONIC_URI_PREFIX = "https://subsonic.org/entity/song/"


def _serialize_to_jspf(playlist, created_for=None, track_count=None):
    """
        Serialize a playlist to JSPF.

        Arguments:
            created_for: The user name of the user for whom this playlist
                         was created.
            track_count: The number of tracks to serialize. If not provided,
                         all tracks will be serialized.
    """

    data = {"creator": "ListenBrainz Troi", "extension": {PLAYLIST_EXTENSION_URI: {"public": True}}}

    if playlist.name:
        data["title"] = playlist.name

    if playlist.description:
        data["annotation"] = playlist.description

    if created_for:
        data["extension"][PLAYLIST_EXTENSION_URI]["created_for"] = created_for

    if playlist.additional_metadata:
        data["extension"][PLAYLIST_EXTENSION_URI]["additional_metadata"] = playlist.additional_metadata

    if not track_count or track_count < 0 or track_count > len(playlist.recordings):
        track_count = len(playlist.recordings)

    tracks = []
    for e in playlist.recordings[:track_count]:
        track = {}
        artist_mbids = []
        if e.artist_credit is not None:
            artist_mbids = [str(artist.mbid) for artist in e.artist_credit.artists or []]
            track["creator"] = e.artist_credit.name if e.artist_credit else ""

        track["title"] = e.name
        track["identifier"] = ["https://musicbrainz.org/recording/" + str(e.mbid)]

        loc = e.musicbrainz.get("filename", None)
        if loc is not None:
            track["location"] = loc

        if e.duration is not None:
            track["duration"] = e.duration

        if artist_mbids:
            track["extension"] = {
                PLAYLIST_TRACK_EXTENSION_URI: {
                    "artist_identifiers": artist_mbids,
                }
            }

        if e.release is not None:
            if e.release.name is not None and e.release.name != "":
                track["album"] = e.release.name

            if e.release.mbid is not None and e.release.mbid != "":
                track["extension"][PLAYLIST_TRACK_EXTENSION_URI]["release_identifier"] = \
                      PLAYLIST_RELEASE_URI_PREFIX + e.release.mbid

        # Output subsonic_ids to the playlist
        subsonic_id = e.musicbrainz.get("subsonic_id", None)
        if subsonic_id is not None:
            if "additional_metadata" not in track["extension"][PLAYLIST_TRACK_EXTENSION_URI]:
                track["extension"][PLAYLIST_TRACK_EXTENSION_URI]["additional_metadata"] = {}
            track["extension"][PLAYLIST_TRACK_EXTENSION_URI]["additional_metadata"]["subsonic_identifier"] = \
                  SUBSONIC_URI_PREFIX + subsonic_id

        tracks.append(track)

    data['track'] = tracks

    return {"playlist": data}


def _deserialize_from_jspf(data) -> Playlist:
    """ Deserialize a ListenBrainz JSPF playlist to a Playlist entity. """
    data = data["playlist"]
    recordings = []

    for track in data["track"]:
        identifiers = track["identifier"]
        if isinstance(identifiers, str):
            identifiers = [identifiers]

        mbid = None
        for identifier in identifiers:
            if identifier.startswith("https://musicbrainz.org/recording/") or \
                    identifier.startswith("http://musicbrainz.org/recording/"):
                mbid = identifier.split("/")[-1]
                break

        recording = Recording(name=track["title"], mbid=mbid)
        if track.get("creator"):
            extension = track["extension"][PLAYLIST_TRACK_EXTENSION_URI]
            if extension.get("artist_identifiers"):
                artist_mbids = [url.split("/")[-1] for url in extension.get("artist_identifiers")]
                artists = [Artist(mbid=mbid) for mbid in artist_mbids]
            else:
                artists = None
            recording.artist_credit = ArtistCredit(name=track["creator"], artists=artists)

        if track.get("album"):
            recording.release = Release(name=track["album"])

        musicbrainz = {}
        if track.get("location"):
            musicbrainz["filename"] = track.get("location")

        try:
            musicbrainz["subsonic_id"] = track["extension"][PLAYLIST_TRACK_EXTENSION_URI] \
                                              ["additional_metadata"]["subsonic_identifier"]
            musicbrainz["subsonic_id"] = musicbrainz["subsonic_id"][len(SUBSONIC_URI_PREFIX):]
        except KeyError:
            pass

        recording.musicbrainz = musicbrainz
        recordings.append(recording)

    try:
        ident = data["identifier"].split("/")[-1],
    except KeyError:
        ident = ""
    playlist = Playlist(name=data["title"],
                        description=data.get("annotation"),
                        mbid=ident,
                        recordings=recordings)
    return playlist


class PlaylistElement(Element):
    """
        Take a list of Recordings or Playlists and save, print or submit them. The playlist element takes
        lists of recordings or lists of playlists and saves them to disk or submits them to ListenBrainz.

        This element can receive a list of recordings in which case that will be treated as a single
        playlist. This element can also recevied a list of playlists, in which case each playlist will
        be saved/submitted individually.

        Note that the Playlist entity object is distinct from this class -- the Playlist object tracks
        the core components of a playlist for passing through troi piplines, whereas this clas
        is designed to be the end of the pipeline for saving results.
    """

    def __init__(self):
        super().__init__()
        self.playlists = []
        self.print_recording = PrintRecordingList()

    @staticmethod
    def inputs():
        return [Recording, Playlist]

    def __str__(self):
        return str(self.playlists)

    def read(self, inputs):

        for input in inputs:
            if len(input) == 0:
                logger.info("No recordings or playlists generated to save.")
                continue

            if isinstance(input[0], Recording):
                if not is_homogeneous(input):
                    raise TypeError("entity list not homogeneous")
                self.playlists.append(Playlist(recordings=input))
            elif isinstance(input[0], Playlist):
                if not is_homogeneous(input):
                    raise TypeError("entity list not homogeneous")
                self.playlists.extend(input)
            else:
                raise PipelineError("Playlist passed incorrect input types.")

        return inputs[0]

    def print(self):
        """Prints the resultant playlists, one after another."""

        if not self.playlists:
            logger.error("[no playlist(s) generated yet]")
            return

        for i, playlist in enumerate(self.playlists):
            if playlist.name:
                logger.info("playlist: '%s'" % playlist.name)
            else:
                logger.info("playlist: %d" % i)

            for recording in playlist.recordings:
                if not recording:
                    logger.info("[invalid Recording]")
                    continue
                self.print_recording.print(recording)

            if playlist.description:
                logger.info("description: '%s'" % playlist.description)

    def save(self, track_count=None, file_obj=None):
        """Save each playlist to disk, giving each playlist a unique name if none was provided.

           Arguments:
              track_count: If provided, write out only this many tracks to the playlist
              file_obj: If provided, write the JSPF file to this file_object
        """

        if not self.playlists:
            raise PipelineError("Playlists have not been generated yet.")

        for i, playlist in enumerate(self.playlists):
            if not file_obj:
                filename = playlist.filename or "playlist_%03d.jspf" % i
                with open(filename, "w") as f:
                    f.write(json.dumps(_serialize_to_jspf(playlist, track_count=track_count)))
            else:
                file_obj.write(json.dumps(_serialize_to_jspf(playlist, track_count=track_count)))

    def get_jspf(self, playlist_index=0):
        """Get the JSPF for the playlist at the given playlist index. Defaults to index 0."""

        try:
            return _serialize_to_jspf(self.playlists[playlist_index])
        except IndexError:
            return None

    def submit(self, token, created_for=None):
        """
            Submit the playlist to ListenBrainz.

            :param token: the ListenBrainz user token to use to submit this playlist.
            :param created_for: the ListenBrainz user name for whom this playlist was created. the token above must be an Approved Playlist Bot in the ListenBrainz server, otherwise the subission will fail.
        """

        if not self.playlists:
            raise PipelineError("Playlists have not been generated yet.")

        playlist_mbids = []
        for playlist in self.playlists:
            if len(playlist.recordings) == 0:
                continue

            logger.info("submit %d tracks" % len(playlist.recordings))
            if playlist.patch_slug is not None:
                playlist.add_metadata({"algorithm_metadata": {"source_patch": playlist.patch_slug}})

            while True:
                r = requests.post(LISTENBRAINZ_PLAYLIST_CREATE_URL,
                                  json=_serialize_to_jspf(playlist, created_for),
                                  headers={"Authorization": "Token " + str(token)})
                if r.status_code == 429:
                    sleep(2)
                    continue

                if r.status_code != 200:
                    try:
                        err = r.json()["error"]
                    except json.decoder.JSONDecodeError:
                        err = r.text

                    raise PipelineError("Cannot post playlist to ListenBrainz: HTTP code %d: %s" % (r.status_code, err))

                break

            try:
                result = json.loads(r.text)
            except ValueError as err:
                raise PipelineError("Cannot post playlist to ListenBrainz: " + str(err))

            playlist.mbid = result["playlist_mbid"]
            playlist_mbids.append((LISTENBRAINZ_SERVER_URL + "/playlist/" + result["playlist_mbid"], result["playlist_mbid"]))

        return playlist_mbids

    def submit_to_spotify(self,
                          user_id: str,
                          token: str,
                          is_public: bool = True,
                          is_collaborative: bool = False,
                          existing_urls: str = None):
        """ Given spotify user id, spotify auth token with appropriate permissions and playlist visibility
         characteristics, upload the playlists generated in the current element to Spotify and return the
         urls of submitted playlists.

        If existing urls are specified then is_public and is_collaborative arguments are ignored.
        """
        sp = spotipy.Spotify(auth=token)
        submitted = []
        for idx, playlist in enumerate(self.playlists):
            if len(playlist.recordings) == 0:
                continue

            existing_url = None
            if existing_urls and idx < len(existing_urls) and existing_urls[idx]:
                existing_url = existing_urls[idx]

            playlist_url, playlist_id = submit_to_spotify(sp, playlist, user_id, is_public, is_collaborative, existing_url)
            submitted.append((playlist_url, playlist_id))

        return submitted

    def submit_to_apple_music(self,
                          user_token: str,
                          developer_token: str,
                          is_public: bool = True,
                          existing_urls: str = None):
        """ Given apple music user token, developer token, upload the playlists generated in the current element to Apple Music and return the
        urls of submitted playlists.

        """
        apple = AppleMusicAPI(developer_token, user_token)
        submitted = []
        for idx, playlist in enumerate(self.playlists):
            if len(playlist.recordings) == 0:
                continue

            existing_url = None
            if existing_urls and idx < len(existing_urls) and existing_urls[idx]:
                existing_url = existing_urls[idx]

            playlist_url, playlist_id = submit_to_apple_music(apple, playlist, is_public, existing_url)
            submitted.append((playlist_url, playlist_id))

        return submitted

    def submit_to_soundcloud(self,
                             user_id: str,
                             access_token: str,
                             is_public: bool = True,
                             existing_urls: str = None):
        """ Given soundcloud user id, soundcloud auth token, upload the playlists generated in the current element to Soundcloud and return the
        urls of submitted playlists.

        """
        sd = SoundcloudAPI(access_token=access_token)
        submitted = []

        for idx, playlist in enumerate(self.playlists):
            if len(playlist.recordings) == 0:
                continue

            existing_url = None
            if existing_urls and idx < len(existing_urls) and existing_urls[idx]:
                existing_url = existing_urls[idx]

            playlist_url, playlist_id = submit_to_soundcloud(sd, playlist, is_public, existing_url)
            submitted.append((playlist_url, playlist_id))

        return submitted


class PlaylistRedundancyReducerElement(Element):

    '''
        This element takes a larger playlist and whittles it down to a smaller playlist by
        removing some tracks in order to reduce the number of times a single artist appears
        in the playlist.
    '''

    def __init__(self, max_artist_occurrence=2, max_num_recordings=50):
        super().__init__()
        self.max_num_recordings = max_num_recordings
        self.max_artist_occurrence = max_artist_occurrence

    @staticmethod
    def inputs():
        return [Playlist]

    @staticmethod
    def outputs():
        return [Playlist]

    def read(self, inputs):

        playlists = []

        for playlist in inputs[0]:
            kept = []
            artists = defaultdict(int)
            for r in playlist.recordings:
                keep = True
                for mbid in r.artist.mbids:
                    if artists[mbid] >= self.max_artist_occurrence:
                        keep = False
                        break
                    artists[mbid] += 1

                if keep:
                    kept.append(r)

            playlist.recordings = kept[:self.max_num_recordings]

        return inputs[0]


class PlaylistShuffleElement(Element):
    '''
        Take in a list of playlists, pass on shuffled playlists.
    '''

    def __init__(self):
        super().__init__()

    @staticmethod
    def inputs():
        return [Playlist]

    @staticmethod
    def outputs():
        return [Playlist]

    def read(self, inputs):

        for playlist in inputs[0]:
            assert type(playlist) == Playlist
            playlist.shuffle()

        return inputs[0]


class PlaylistBPMSawtoothSortElement(Element):
    '''
        Sort a playlist by BPM going from slow to fast and back to slow.
    '''

    def __init__(self):
        super().__init__()

    @staticmethod
    def inputs():
        return [Playlist]

    @staticmethod
    def outputs():
        return [Playlist]

    def bpm_sawtooth_sort(self, recordings):
        try:
            sorted_recs = sorted(recordings, key=lambda rec: rec.acousticbrainz["bpm"])
        except AttributeError:
            raise RuntimeError("acousticbrainz.bpm not set for recording in playlist PBM sort.")

        # Sort the recordings by BPM ASC. Then, walking from the back of the array to the front,
        # Move every second track to the back of the array, creating a sawtooth shape of BPMs
        index = sorted_recs.index(max(sorted_recs, key=lambda rec: rec.acousticbrainz["bpm"]))
        while index >= 0:
            sorted_recs.append(sorted_recs.pop(index))
            index -= 2

        return sorted_recs

    def read(self, inputs):
        for playlist in inputs[0]:
            playlist.recordings = self.bpm_sawtooth_sort(playlist.recordings)

        return inputs[0]


class PlaylistMakerElement(Element):
    '''
        This element takes in Recordings and spits out a Playlist.

        :param name: The name of the playlist
        :param desc: The description of the playlist
        :param patch-slug: The patch slug (URL-safe short name) of the patch that created the playlist. Optional.
        :param max_num_recordings: The maximum number of recordings to have in the playlist. Extras are discarded. Optional argument, and the default is to keep all recordings
        :param max_artist_occurrence: The number of times and artist is allowed to appear in the playlist. Any recordings that exceed this count ared discarded. Optional, default is to keep all recordings.
        :param shuffle: If True, the playlist will be shuffled before being truncated. Optional. Default: False
        :param is_april_first: If True, do something very sneaky. Default: False.
    '''

    def __init__(self,
                 name=None,
                 desc=None,
                 patch_slug=None,
                 user_name=None,
                 max_num_recordings=None,
                 max_artist_occurrence=None,
                 shuffle=False,
                 expires_at=None,
                 is_april_first=False):
        super().__init__()
        self.name = name
        self.desc = desc
        self.patch_slug = patch_slug
        self.user_name = user_name
        self.max_num_recordings = max_num_recordings
        self.max_artist_occurrence = max_artist_occurrence
        self.shuffle = shuffle
        self.expires_at = expires_at
        self.is_april_first = is_april_first

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        # We never actually return more than one playlist.
        return [Playlist]

    def read(self, inputs):
        recordings = inputs[0]

        if self.max_num_recordings is None:
            max_num_recordings = len(recordings)
        else:
            max_num_recordings = self.max_num_recordings

        if self.max_artist_occurrence is not None:
            kept = []
            artists = defaultdict(int)
            for r in recordings:
                keep = True
                for mbid in [ a.mbid for a in r.artist_credit.artists ]:
                    if artists[mbid] >= self.max_artist_occurrence:
                        keep = False
                        break
                    artists[mbid] += 1

                if keep:
                    kept.append(r)

            recordings = kept[:max_num_recordings]
        else:
            recordings = recordings[:max_num_recordings]

        # Call the patch's post_process function
        self.patch.post_process()

        if self.name is None:
            self.name = self.local_storage["_playlist_name"]

        if self.desc is None:
            self.desc = self.local_storage["_playlist_desc"]

        playlist = Playlist(name=self.name,
                            description=self.desc,
                            recordings=recordings,
                            patch_slug=self.patch_slug,
                            user_name=self.user_name)
        if self.shuffle:
            playlist.shuffle()

        if self.expires_at is not None:
            playlist.add_metadata({ "expires_at": self.expires_at.isoformat() })

        if self.is_april_first:
            try:
                playlist.recordings[2].mbid = "8f3471b5-7e6a-48da-86a9-c1c07a0f47ae"
            except IndexError:
                pass

        return [playlist]


class RecordingsFromMusicServiceElement(Element):
    """ Create a troi.Playlist entity from track and artist names."""

    def __init__(self, token=None, playlist_id=None, music_service=None, apple_user_token=None):
        """
            Args:
                playlist_id: id of the playlist to be used for creating the playlist element
                token: the Spotify token to fetch the playlist tracks
                music_service: the name of the music service to be used for fetching the playlist data
                apple_music_token (optional): the user token for Apple Music API 
        """
        super().__init__()
        self.token = token
        self.playlist_id = playlist_id
        self.music_service = music_service
        self.apple_user_token = apple_user_token


    @staticmethod
    def outputs():
        return [ Recording ]

    def read(self, inputs):
        recordings = []

        mbid_mapped_tracks = music_service_tracks_to_mbid(self.token, self.playlist_id, self.music_service, self.apple_user_token)
        if mbid_mapped_tracks:
            for mbid in mbid_mapped_tracks:
                recordings.append(Recording(mbid=mbid))

        return recordings

    
class PlaylistFromJSPFElement(Element):
    """ Create a troi.Playlist entity from a ListenBrainz JSPF playlist or LB playlist."""

    def __init__(self, playlist_mbid=None, jspf=None, token=None):
        """
            The caller must pass either playlist_mbid or the jspf itself, but not both.
            Args:
                playlist_mbid: mbid of the ListenBrainz playlist to be used for creating the playlist element
                jspf: The actual JSPF for the playlist.
                token: the listenbrainz auth token to fetch the playlist, only needed for private playlists
        """
        super().__init__()
        self.playlist_mbid = playlist_mbid
        self.token = token
        self.jspf = jspf

        if self.jspf is not None and self.playlist_mbid is not None:
            raise RuntimeError("Pass either jspf or playlist_mbid to PlaylistFromJSPFElement, not both.")

    @staticmethod
    def outputs():
        return [Playlist]

    def read(self, inputs):
        if self.playlist_mbid:
            headers = None
            if self.token:
                headers = {"Authorization": f"Token {self.token}"}
            response = requests.get(LISTENBRAINZ_PLAYLIST_FETCH_URL + self.playlist_mbid, headers=headers)
            response.raise_for_status()
            data = response.json()
            return [_deserialize_from_jspf(data)]
        else:
            return [_deserialize_from_jspf(self.jspf)]


class DumpElement(Element):
    """
        Accept whatever and print it out in a reasonably sane manner.
    """

    def __init__(self):
        super().__init__()

    @staticmethod
    def inputs():
        return [Recording, Playlist]

    @staticmethod
    def outputs():
        return [Recording, Playlist]

    def read(self, inputs):
        for input in inputs:
            pr = PrintRecordingList()
            pr.print(input)

        return inputs[0]



import os
from collections import defaultdict
import datetime
import sys

import peewee
import requests

from troi.content_resolver.model.database import db
from troi.content_resolver.model.recording import Recording, RecordingMetadata
from troi.content_resolver.utils import select_recordings_on_popularity
from troi.recording_search_service import RecordingSearchByTagService
from troi.plist import plist


class LocalRecordingSearchByTagService(RecordingSearchByTagService):
    '''
    Given the local database, search for recordings that meet given tag criteria
    '''

    def __init__(self):
        RecordingSearchByTagService.__init__(self)

    def search(self, tags, operator, pop_begin, pop_end, num_recordings):
        """
        Perform a tag search. Parameters:

        tags - a list of string tags to search for
        operator - a string specifying "or" or "and"
        pop_begin - if many recordings match the above parameters, return only
                        recordings that have a minimum popularity percent score
                        of pop_begin.
        pop_end - if many recordings match the above parameters, return only
                      recordings that have a maximum popularity percent score
                      of pop_end.

        If only few recordings match, the pop_begin and pop_end are
        ignored.
        """

        # Search for all recordings that match the given tags with given operator
        if operator == "or":
            query, params, pop_clause = self.or_search(tags)
        else:
            query, params, pop_clause = self.and_search(tags)

        placeholders = ",".join(("?", ) * len(tags))
        cursor = db.execute_sql(query % (placeholders, pop_clause), params)

        recordings = []
        for rec in cursor.fetchall():
            recordings.append({"recording_mbid": rec[0], "popularity": rec[1], "file_id": rec[2], "file_id_type": rec[3]})

        return select_recordings_on_popularity(recordings, pop_begin, pop_end, num_recordings)

    def or_search(self, tags, min_popularity=None, max_popularity=None):
        """
            Return the sql query that finds recordings using the OR operator
        """

        query = """WITH recording_ids AS (
                        SELECT DISTINCT(recording_id)
                          FROM tag
                          JOIN recording_tag
                            ON recording_tag.tag_id = tag.id
                          JOIN recording
                            ON recording.id = recording_tag.recording_id
                         WHERE name in (%s)
                   )
                       SELECT recording_mbid
                            , popularity AS percent
                            , file_id
                            , file_id_type
                         FROM recording
                         JOIN recording_ids
                           ON recording.id = recording_ids.recording_id
                         JOIN recording_metadata
                           ON recording.id = recording_metadata.recording_id
                           %s
                     ORDER BY popularity DESC"""

        if min_popularity is not None and max_popularity is not None:
            pop_clause = """WHERE popularity >= %.4f AND popularity < %.4f""" % \
                (min_popularity, max_popularity)
        else:
            pop_clause = ""

        return query, tags, pop_clause

    def and_search(self, tags, min_popularity=None, max_popularity=None):
        """
            Return the sql query that finds recordings using the AND operator
        """
        query = """WITH recording_tags AS (
                        SELECT DISTINCT recording.id AS recording_id
                             , tag.name AS tag_name
                          FROM tag
                          JOIN recording_tag
                            ON recording_tag.tag_id = tag.id
                          JOIN recording
                            ON recording.id = recording_tag.recording_id
                         WHERE name in (%s)
                         ORDER BY recording.id
                   ), recording_ids AS (
                       SELECT recording_tags.recording_id
                         FROM recording_tags
                         JOIN recording_metadata
                           ON recording_tags.recording_id = recording_metadata.recording_id
                     GROUP BY recording_tags.recording_id
                       HAVING count(recording_tags.tag_name) = ?
                   )
                       SELECT recording_mbid
                            , popularity AS percent
                            , file_id
                            , file_id_type
                         FROM recording
                         JOIN recording_ids
                           ON recording.id = recording_ids.recording_id
                         JOIN recording_metadata
                           ON recording.id = recording_metadata.recording_id
                           %s
                     ORDER BY popularity DESC"""

        if min_popularity is not None and max_popularity is not None:
            pop_clause = """WHERE popularity >= %.4f AND popularity < %.4f""" % \
                (min_popularity, max_popularity)
        else:
            pop_clause = ""

        return query, (*tags, len(tags)), pop_clause



import requests

from troi import Element, Artist, Recording, Playlist, PipelineError


class DataSetFetcherElement(Element):
    '''
        Fetch a dataset from the ListenBrainz experimental APIs that contains at least a recording_mbid.
        Recording.name, listenbrainz.listen_count and ArtistCredit.name will also be filled out
        if it is available in the returned data.

        :param server_url: the URL to POST to
        :param json_post_data: the dict data that will be POSTed as JSON
    '''

    def __init__(self, server_url, json_post_data, max_num_items=None):
        super().__init__()
        self.server_url = server_url
        self.json_post_data = json_post_data
        self.max_num_items = max_num_items

    @staticmethod
    def inputs():
        return []

    @staticmethod
    def outputs():
        return [Recording, Artist]

    def create_recording(self, row):

        r.artist = self.create_artist(row)

        if 'recording_name' in row:
            r.name = row['recording_name']

        if 'year' in row:
            r.year = row['year']

        if 'listen_count' in row:
            r.listenbrainz = {"listen_count": row["listen_count"]}

        if 'bpm' in row:
            r.acousticbrainz = {"bpm": row["bpm"]}

        return r

    def create_artist(self, row):

        if 'artist_mbid' in row:
            return Artist(mbids=[row['artist_mbid']])

        if 'artist_credit_name' in row:
            return Artist(name=row['artist_credit_name'])

        if 'artist_credit_id' in row:
            return Artist(artist_credit_id=row['artist_credit_id'])

        if 'artist_mbids' in row:
            return Artist(mbids=row['artist_mbids'])

        return None

    def read(self, inputs):

        r = requests.post(self.server_url, json=self.json_post_data)
        if r.status_code != 200:
            raise PipelineError("Cannot fetch  dataset recordings from %s. HTTP code %s (%s)" %
                                (self.server_url, r.status_code, r.text))

        try:
            data = r.json()[3]["data"]
        except (IndexError, KeyError):
            data = r.json()

        output = []
        for row in data:
            if "recordind_mbid" in row:
                r = self.create_recording(row)
            elif "artist_mbid" in row:
                r = self.create_artist(row)
            else:
                continue

        r = Recording(mbid=row['recording_mbid'])
        if 'artist_credit_name' in row:
            if 'artist_credit_id' in row:
                r.artist_credit = ArtistCredit(artist_credit_id = row['artist_credit_id'],
                                               name = row['artist_credit_name'])
            output.append(r)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'troi'
copyright = '2020, MetaBrainz Foundation'
author = 'MetaBrainz Foundation'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinxcontrib.autohttp.flask',
    'sphinx_click',
    'sphinx.ext.autosectionlabel',
]
autosectionlabel_prefix_document = True

# General information about the project.
project = u'Troi'
copyright = u'2017-2018, MetaBrainz Foundation'
author = u'MetaBrainz Foundation'

# The short X.Y version.
version = '0.1.0'
# The full version, including alpha/beta/rc tags.
release = '0.1.0'

language = "en"
pygments_style = 'sphinx'



# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
import sphinx_rtd_theme
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'canonical_url': "https://listenbrainz.readthedocs.io/en/production/",
    'collapse_navigation': False,
    'display_version': False,
    'navigation_depth': 3,
}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_short_title = "ListenBrainz Troi docs"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "listenbrainz-logo.svg"



# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



import datetime
import logging
from collections import defaultdict
from operator import itemgetter
from time import sleep

import requests

from troi.content_resolver.model.database import db

logger = logging.getLogger(__name__)


class UnresolvedRecordingTracker:
    '''
        This class keeps track of recordings that were not resolved when
        a playlist was resolved. This will allow us to give recommendations
        on which albums to add to their collection to resolve more recordings.
    '''

    LOOKUP_BATCH_SIZE = 50

    def __init__(self):
        pass

    @staticmethod
    def chunks(lst, n):
        """Yield successive n-sized chunks from lst."""
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    @staticmethod
    def multisort(xs, specs):
        """ Multiple key sort helper """
        for key, reverse in reversed(specs):
            xs.sort(key=itemgetter(key), reverse=reverse)
        return xs

    def add(self, recording_mbids):
        """
            Add one or more recording MBIDs to the unresolved recordings track. If this has
            previously been unresolved, increment the count for the number
            of times it has been unresolved.
        """

        recording_mbids = tuple(set(recording_mbids))

        placeholders = ",".join(("?", ) * len(recording_mbids))
        existing = {
                row[0]:True for row in db.execute_sql(
                """SELECT recording_mbid
                     FROM unresolved_recording
                    WHERE recording_mbid IN (%s)""" %
                placeholders, recording_mbids).fetchall()
        }

        now = datetime.datetime.now()
        with db.atomic():
            for mbid in recording_mbids:
                if mbid in existing:
                    db.execute_sql("""UPDATE unresolved_recording
                                         SET lookup_count = lookup_count + 1,
                                             last_updated = ?
                                       WHERE recording_mbid = ?""", (now, mbid))
                else:
                    db.execute_sql("""INSERT INTO unresolved_recording (recording_mbid, last_updated, lookup_count)
                                           VALUES (?, ?, ?)""", (mbid, datetime.datetime.now(), 1))


        # For when UPSERT is available on RPi
        #query = """INSERT INTO unresolved_recording (recording_mbid, last_updated, lookup_count)
        #                VALUES (?, ?, 1)
        # ON CONFLICT DO UPDATE SET lookup_count = EXCLUDED.lookup_count + 1"""

    def get_releases(self):
        """
            Organize the unresolved recordings into releases with a list of recordings.
            Return up to num_item releases.
        """

        # First call cleanup, which removes recordings that may have been recently
        # added
        self.cleanup()

        query = f"""SELECT recording_mbid
                         , lookup_count
                      FROM unresolved_recording"""

        cursor = db.execute_sql(query)
        recording_mbids = []
        lookup_counts = {}
        for row in cursor.fetchall():
            recording_mbids.append(row[0])
            lookup_counts[row[0]] = row[1]

        recording_data = {}
        for chunk in self.chunks(recording_mbids, self.LOOKUP_BATCH_SIZE):
            args = ",".join(chunk)

            params = {"recording_mbids": args, "inc": "artist release"}
            while True:
                r = requests.get("https://api.listenbrainz.org/1/metadata/recording", params=params)
                if r.status_code == 429:
                    sleep(1)
                    continue

                if r.status_code != 200:
                    logger.info("Failed to fetch metadata for recordings: ", r.text)
                    return []

                break
            recording_data.update(dict(r.json()))

        releases = defaultdict(list)
        for mbid in recording_mbids:
            rec = recording_data[mbid]
            releases[rec["release"]["mbid"]].append({
                "artist_name": rec["artist"]["name"],
                "artists": rec["artist"]["artists"],
                "release_name": rec["release"]["name"],
                "release_mbid": rec["release"]["mbid"],
                "release_group_mbid": rec["release"]["release_group_mbid"],
                "recording_name": rec["recording"]["name"],
                "recording_mbid": mbid,
                "lookup_count": lookup_counts[mbid]
            })

        release_list = []
        for mbid in releases:
            release = releases[mbid]
            total_count = sum([rec["lookup_count"] for rec in release])
            release_list.append({
                "mbid": release[0]["release_mbid"],
                "release_name": release[0]["release_name"],
                "artist_name": release[0]["artist_name"],
                "lookup_count": total_count,
                "recordings": release
            })

        return self.multisort(release_list, (("lookup_count", True), ("artist_name", False), ("release_name", False)))

    def print_releases(self, releases):
        """ Neatly print all the release/recordings returned from the get_releases function """

        logger.info("%-60s %-50s" % ("RELEASE", "ARTIST"))
        for release in releases:
            logger.info("%-60s %-50s" % (release["release_name"][:59], release["artist_name"][:49]))
            for rec in release["recordings"]:
                logger.info("   %-57s %d lookups" % (rec["recording_name"][:56], rec["lookup_count"]))
            logger.info("")

    def cleanup(self):
        """
            Check the local collection and remove any recordings from unresolved recordings that have
            been added to the DB.
        """

        query = f"""DELETE FROM unresolved_recording
                          WHERE recording_mbid in (
                                 SELECT recording.recording_mbid
                                   FROM recording
                                   JOIN unresolved_recording
                                     ON recording.recording_mbid = unresolved_recording.recording_mbid
                                 )"""

        cursor = db.execute_sql(query)



import logging

from troi import Element, Artist, ArtistCredit, Release, Recording
import liblistenbrainz
import liblistenbrainz.errors

logger = logging.getLogger(__name__)


class UserArtistsElement(Element):
    '''
        Fetch artist statistics from ListenBrainz.

        :param user_name: The ListenBrainz user for whom to fetch statistics.
        :param time_range: The time range for which to fetch stats.
        :param count: The number of items to return. Default 25.
        :param offset: The offset into the results. Used for pagination.
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time', auth_token=None):
        super().__init__()
        self.client = liblistenbrainz.ListenBrainz()
        if auth_token:
            self.client.set_auth_token(auth_token)
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [ArtistCredit]

    def read(self, inputs=[]):

        artist_list = []
        artists = self.client.get_user_artists(self.user_name, self.count, self.offset, self.time_range)
        artist_credits = []
        for a in artists['payload']['artists']:
            artists = [Artist(mbid=mbid) for mbid in a['artist_mbids']]
            artist_list.append(ArtistCredit(name=a['artist_name'], artists=artists))

        return artist_list


class UserReleasesElement(Element):
    '''
        Fetch release statistics from ListenBrainz

        :param user_name: The ListenBrainz user for whom to fetch statistics.
        :param time_range: The time range for which to fetch stats.
        :param count: The number of items to return. Default 25.
        :param offset: The offset into the results. Used for pagination.
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time', auth_token=None):
        super().__init__()
        self.client = liblistenbrainz.ListenBrainz()
        if auth_token:
            self.client.set_auth_token(auth_token)
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [Release]

    def read(self, inputs=[]):

        release_list = []
        releases = self.client.get_user_releases(self.user_name, self.count, self.offset, self.time_range)
        for r in releases['payload']['releases']:
            artists = [Artist(mbid=mbid) for mbid in r['artist_mbids']]
            release_list.append(
                Release(r['release_name'],
                        mbid=r['release_mbid'],
                        artist_credit=ArtistCredit(name=r['artist_name'], artists=artists)))

        return release_list


class UserRecordingElement(Element):
    '''
        Fetch release statistics from ListenBrainz

        :param user_name: The ListenBrainz user for whom to fetch statistics.
        :param time_range: The time range for which to fetch stats.
        :param count: The number of items to return. Default 25.
        :param offset: The offset into the results. Used for pagination.
    '''

    def __init__(self, user_name, count=25, offset=0, time_range='all_time'):
        super().__init__()
        self.client = liblistenbrainz.ListenBrainz()
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.time_range = time_range

    def outputs(self):
        return [Recording]

    def read(self, inputs=[]):
        recording_list = []
        try:
            recordings = self.client.get_user_recordings(self.user_name, self.count, self.offset, self.time_range)
        except liblistenbrainz.errors.ListenBrainzAPIException as err:
            logger.info("Cannot fetch recording stats for user %s" % self.user_name)
            return []

        if recordings is None or "recordings" not in recordings['payload']:
            return []

        for r in recordings['payload']['recordings']:
            artists = [Artist(mbid=mbid) for mbid in r['artist_mbids']]
            artist_credit = ArtistCredit(r['artist_name'], artists=artists)
            release = Release(r['release_name'], mbid=r['release_mbid'])
            recording_list.append(Recording(r['track_name'], mbid=r['recording_mbid'], artist_credit=artist_credit,
                                            release=release))

        return recording_list



import requests
import ujson

from troi import Element, Recording


class MBIDReaderElement(Element):
    '''
        Read MBIDs from a file, one per line and return Recording objects with the MBID field filled out.

        :param filename: The name (with full path) of the file to read recording MBIDs from.
    '''

    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    @staticmethod
    def inputs():
        return [ ]

    @staticmethod
    def outputs():
        return [ Recording ]

    def read(self, inputs):

        mbids = []
        with open(self.filename, "r") as f:
            for line in f.readlines():
                if not line:
                    break

                mbid = line.strip()
                mbids.append(Recording(mbid=mbid))

        return mbids



import requests
import json
import logging
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)
SOUNDCLOUD_URL = f"https://api.soundcloud.com/"
APPLE_MUSIC_URL = f"https://api.music.apple.com/v1"

class AppleMusicException(Exception):
    def __init__(self, code, msg):
        self.code = code
        self.msg = msg
        super().__init__(f"Apple Music api error {code}: {msg}")


class AppleMusicAPI:
    def __init__(self, developer_token, user_token):
        self.developer_token = developer_token
        self.user_token = user_token
        self.headers = {
            "Authorization": f"Bearer {self.developer_token}",
            "Music-User-Token": self.user_token,
            "Content-Type": "application/json"
        }
        self.session = create_http_session()
        self.storefront = self._get_user_storefront()

    def _get_user_storefront(self):
        """ Fetch a storefront for a specific user.
        """
        url=f"{APPLE_MUSIC_URL}/me/storefront"
        response = self.session.get(url, headers=self.headers)

        data = response.json()["data"][0]["id"]
        return data

    def create_playlist(self, name, is_public=True,description=None):
        url = f"{APPLE_MUSIC_URL}/me/library/playlists"
        data = {
            "attributes": {
                "name": name,
                "isPublic": is_public
            }
        }
        if description:
            data["attributes"]["description"] = description
        response = self.session.post(url, headers=self.headers, data=json.dumps(data))
        return response.json()

    def playlist_add_tracks(self, playlist_id, track_ids):
        """ Adds tracks to a playlist in Apple Music, does not return response
        """
        url = f"{APPLE_MUSIC_URL}/me/library/playlists/{playlist_id}/tracks"
        data = {
            "data": [{"id": track_id, "type": "songs"} for track_id in track_ids]
        }
        self.session.post(url, headers=self.headers, data=json.dumps(data))

    def get_playlist_tracks(self, playlist_id):
        url = f"{APPLE_MUSIC_URL}/me/library/playlists/{playlist_id}?include=tracks"
        response = self.session.get(url, headers=self.headers).json()
        tracks = response["data"][0]["relationships"]["tracks"]["data"]
        total_tracks_count = response["data"][0]["relationships"]["tracks"]["meta"]["total"]
        playlist_name = response["data"][0]["attributes"]["name"]
        playlist_description = response["data"][0]["attributes"].get(
            "description", {}).get("standard", "")

        # apple music returns only the first 100 tracks with "/playlists/{playlist_id}?include=tracks"
        # and we need to fetch the rest of the tracks using the "playlists/{playlist_id}/tracks" endpoint
        if len(tracks) < total_tracks_count:
            offset = len(tracks)
            # endpoint returns 100 tracks per call max -> run iteratively with an offset until there are no more tracks
            while True:
                url = f"{APPLE_MUSIC_URL}/me/library/playlists/{playlist_id}/tracks?limit=100&offset={offset}"
                response = self.session.get(url, headers=self.headers).json()
                try:
                    if "errors" in response:
                        # https: // developer.apple.com/documentation/applemusicapi/errorsresponse
                        error_objects = response["errors"]
                        for error_object in error_objects:
                            error_message = error_object["detail"] if "detail" in error_object else error_object["title"]
                            logger.error(
                                f"Error code {error_object['status']}: {error_message}")
                        break

                    tracks_data = response["data"]
                    tracks.extend(tracks_data) if tracks_data else None

                    if len(tracks_data) == 0 or len(tracks) == total_tracks_count:
                        break
                    # set new offset for the next loop
                    offset = offset + len(tracks_data)
                except KeyError as err:
                    # No data returned from API call, could be an error, do nothing
                    logger.error(
                        "Failed to fetch Apple playlist tracks: %s, ignoring error..." % err)
                    break

        return tracks, playlist_name, playlist_description


class SoundCloudException(Exception):
    def __init__(self, code, msg):
        self.code = code
        self.msg = msg
        super().__init__(f"http error {code}: {msg}")


class SoundcloudAPI:
    def __init__(self, access_token):
        self.access_token = access_token
        self.headers = {
            "Authorization": f"OAuth {self.access_token}",
            "Content-Type": "application/json"
        }
        self.session = create_http_session()

    def create_playlist(self, title, sharing="public", track_ids=None, description=None):
        url = f"{SOUNDCLOUD_URL}/playlists"
        data = {
            "playlist": {
                "title": title,
                "sharing": sharing,
            }
        }
        if track_ids:
            data["playlist"]["tracks"] = [{"id": track_id} for track_id in track_ids]

        if description:
            data["playlist"]["description"] = description
        response = self.session.post(url, headers=self.headers, data=json.dumps(data))
        return response.json()

    def add_playlist_tracks(self, playlist_id, track_ids):
        url = f"{SOUNDCLOUD_URL}/playlists/{playlist_id}"
        data = {
            "playlist": {
                "tracks": [{"id": track_id} for track_id in track_ids]
            }
        }
        response = self.session.put(url, headers=self.headers, data=json.dumps(data))
        return response.json()

    def update_playlist_details(self, playlist_id, title=None, description=None, track_ids=None):
        url = f"{SOUNDCLOUD_URL}/playlists/{playlist_id}"
        data = {
            "playlist": {}
        }
        if title:
            data["playlist"]["title"] = title
        if description:
            data["playlist"]["description"] = description
        if track_ids:
            data["playlist"]["tracks"] = [{"id": track_id} for track_id in track_ids]

        response = self.session.put(url, headers=self.headers, data=json.dumps(list(data)))
        return response.json()

    def get_track_details(self, track_ids):
        track_details = []

        for track_id in track_ids:
            url = f"{SOUNDCLOUD_URL}/tracks/{track_id}"
            response = self.session.get(url, headers=self.headers)
            data = response.json()
            track_details.append(data)

        return track_details


    def get_playlist(self, playlist_id):
        url = f"{SOUNDCLOUD_URL}/playlists/{playlist_id}"

        response = self.session.get(url, headers=self.headers)
        return response.json()


    def get_playlist_tracks(self, playlist_id, **kwargs):
        tracks = []
        url = f"{SOUNDCLOUD_URL}/playlists/{playlist_id}/tracks"

        while url:
            response = self.session.get(url, headers=self.headers, params=kwargs)
            data = response.json()
            tracks.extend(data.get('collection', []))
            url = data.get("next_href")

        return tracks


def create_http_session():
    """ Create an HTTP session with retry strategy for handling rate limits and server errors.
    """
    retry_strategy = Retry(
        total=3,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "OPTIONS"],
        backoff_factor=1
    )

    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http



from troi import Element

from troi.content_resolver.content_resolver import ContentResolver
from troi.content_resolver.model.recording import Recording, FileIdType
from troi.content_resolver.model.database import db
from troi import Recording


class RecordingResolverElement(Element):
    """
        This Troi element takes in a list of recordings, which *must* have artist name and recording
        name set and resolves them to a local collection by using the ContentResolver class
    """

    def __init__(self, match_threshold, quiet):
        """ Match threshold: The value from 0 to 1.0 on how sure a match must be to be accepted.
        """
        Element.__init__(self)
        self.match_threshold = match_threshold
        self.resolve = ContentResolver(quiet)

    @staticmethod
    def inputs():
        return []

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        # Build the fuzzy index
        lookup_data = []
        for recording in inputs[0]:
            if recording.artist_credit is None or recording.artist_credit.name is None or recording.name is None:
                raise RuntimeError("artist name and recording name are needed for RecordingResolverElement.")

            lookup_data.append({"artist_name": recording.artist_credit.name,
                                "recording_name": recording.name,
                                "recording_mbid": recording.mbid})

        self.resolve.build_index()

        # Resolve the recordings
        resolved = self.resolve.resolve_recordings(lookup_data, self.match_threshold)
        recording_ids = tuple([result["recording_id"] for result in resolved])

        if not recording_ids:
            return []

        # Could also be done with, but for some reason it fails when using IN. <shrug>
        # Recording.select().where(Recording.id.in_(recording_ids))

        # Fetch the recordings to lookup file ids
        query = """SELECT recording.id
                        , file_id
                        , file_id_type
                     FROM recording
                    WHERE """

        where_clause_elements = []
        for id in recording_ids:
            where_clause_elements.append("recording.id = %d" % id)

        where_clause = " or ".join(where_clause_elements)
        query += where_clause

        cursor = db.execute_sql(query)
        recordings = []
        for row in cursor.fetchall():
            recordings.append({"recording_id": row[0],
                               "file_id": row[1],
                               "file_id_type": row[2]})

        # Build indexes
        file_id_index = {}
        for recording in recordings:
            file_id_index[recording["recording_id"]] = (recording["file_id"], recording["file_id_type"])

        # Set the ids into the recordings
        results = []
        for r in resolved:
            file_id, file_id_type = file_id_index[r["recording_id"]]
            recording = inputs[0][r["index"]]
            if file_id_type == FileIdType.SUBSONIC_ID.value:
                recording.musicbrainz["subsonic_id"] = file_id
                results.append(recording)

            if file_id_type == FileIdType.FILE_PATH.value:
                recording.musicbrainz["filename"] = file_id
                results.append(recording)

        return results



import requests
from time import sleep
import ujson

from troi import PipelineError, DEVELOPMENT_SERVER_URL

AREA_LOOKUP_SERVER_URL = DEVELOPMENT_SERVER_URL + "/area-lookup/json"
def area_lookup(area_name):
    '''
        Given an area name, lookup the area_id and return it. Return None if area not found.
    '''

    data = [ { '[area]': area_name } ]
    while True:
        r = requests.post(AREA_LOOKUP_SERVER_URL, json=data)
        if r.status_code == 429:
            sleep(2)
            continue
        if r.status_code != 200:
            raise PipelineError("Cannot lookup area name. " + str(r.text))

        break

    try:
        rows = ujson.loads(r.text)
    except ValueError as err:
        raise PipelineError("Cannot lookup area name, invalid JSON returned: " + str(err))

    if len(rows) == 0:
        raise PipelineError("Cannot find area name. Must be spelled exactly as in MusicBrainz.")

    return rows[0]['area_mbid']



import unittest

from troi import Artist, Recording
import troi.sorts 


class TestYearSortElement(unittest.TestCase):

    def test_year_sort_element(self):
        rlist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', year=1980), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', year=1970),
                  Recording(mbid='07316967-cfcb-4920-b8ca-e96dcb4afc08'),
                  Recording(mbid='88b433b3-2c4f-4c8b-839b-77a203ec950f', year=1990) ]

        e = troi.sorts.YearSortElement()
        flist = e.read([rlist])
        assert len(flist) == 4
        assert flist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'
        assert flist[1].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[2].mbid == '88b433b3-2c4f-4c8b-839b-77a203ec950f'
        assert flist[3].mbid == '07316967-cfcb-4920-b8ca-e96dcb4afc08'

        e = troi.sorts.YearSortElement(reverse=True)
        flist = e.read([rlist])
        assert len(flist) == 4
        assert flist[0].mbid == '07316967-cfcb-4920-b8ca-e96dcb4afc08'
        assert flist[1].mbid == '88b433b3-2c4f-4c8b-839b-77a203ec950f'
        assert flist[2].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[3].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'



#!/usr/bin/env python3
import logging
import sys

import click

from troi.logging_utils import set_log_level
from troi.utils import discover_patches
from troi.core import list_patches, patch_info, convert_patch_to_command
from troi.content_resolver.cli import cli as resolver_cli, db_file_check, output_playlist
from troi.content_resolver.subsonic import SubsonicDatabase
from troi.content_resolver.lb_radio import ListenBrainzRadioLocal
from troi.content_resolver.playlist import read_jspf_playlist
from troi.local.periodic_jams_local import PeriodicJamsLocal

logger = logging.getLogger(__name__)

try:
    sys.path.insert(1, ".")
    import config
except ImportError as err:
    config = None


@click.group()
def cli():
    pass


# Add the "db" submenu
resolver_cli.short_help = "database and content resolution commands sub menu"
cli.add_command(resolver_cli, name="db")


@cli.command(context_settings=dict(ignore_unknown_options=True, ))
@click.argument('patch', type=str)
@click.option('--quiet', '-q', 'quiet', help="Do no print out anything", required=False, is_flag=True)
@click.option('--save', '-s', help="Save the generated playlist", required=False, is_flag=True)
@click.option(
    '--token',
    '-t',
    help="Specify the ListenBrainz user token to upload playlist. See https://listenbrainz.org/profile to get your user token.",
    required=False,
    type=click.UUID)
@click.option('--upload',
              '-u',
              help="Upload the generated playlist to ListenBrainz. Also requires --token option",
              required=False,
              is_flag=True)
@click.option('--created-for', '-c', help="The name of the LB user the playlist was created for. For LB use only.", required=False)
@click.option('--name', '-n', help="Override the default name of the generated playlist", required=False)
@click.option('--desc', '-d', help="Override the default description of the generated playlist", required=False)
@click.option('--min-recordings', '-m', help="The minimum number of playlist required for the playlist", type=int, required=False)
@click.option('--spotify-user-id', help="The spotify user name to upload the playlist to", type=str, required=False)
@click.option('--spotify-token',
              help="The spotify token with the correct permissions required to upload playlists",
              type=str,
              required=False)
@click.option('--spotify-url',
              help="instead of creating a new spotify playlist, update the existing playlist at this url",
              type=str,
              required=False,
              multiple=True)
@click.option('--apple-music-developer-token',
              help="The apple music developer token required to upload playlists",
              type=str,
              required=False)
@click.option('--apple-music-user-token',
              help="The apple music user token required to upload playlists on behalf of the user",
              type=str,
              required=False)
@click.option('--apple-music-url',
              help="instead of creating a new apple music playlist, update the existing playlist at this url. currently not supported due to Apple Music API restrictions",
              type=str,
              required=False,
              multiple=True)
@click.option('--soundcloud-user-id', help="The soundcloud user name to upload the playlist to", type=str, required=False)
@click.option('--soundcloud-token',
              help="The soundcloud token with the correct permissions required to upload playlists",
              type=str,
              required=False)
@click.option('--soundcloud-url',
              help="instead of creating a new soundcloud playlist, update the existing playlist at this url",
              type=str,
              required=False,
              multiple=True)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def playlist(patch, quiet, save, token, upload, args, created_for, name, desc, min_recordings, spotify_user_id, spotify_token,
             spotify_url, soundcloud_user_id, soundcloud_token, soundcloud_url,
             apple_music_developer_token, apple_music_user_token, apple_music_url):
    """
    Generate a global MBID based playlist using a patch
    """

    set_log_level(quiet)
    patchname = patch
    patches = discover_patches()
    if patchname not in patches:
        logger.info("Cannot load patch '%s'. Use the list command to get a list of available patches." % patchname)
        return None

    patch_args = {
        "save": save,
        "token": token,
        "created_for": created_for,
        "upload": upload,
        "name": name,
        "desc": desc,
        "min_recordings": min_recordings,
        "quiet": quiet,
    }
    if spotify_token:
        patch_args["spotify"] = {
            "user_id": spotify_user_id,
            "token": spotify_token,
            "is_public": True,
            "is_collaborative": False,
            "existing_urls": spotify_url
        }
    if apple_music_developer_token:
        patch_args["apple_music"] = {
            "developer_token": apple_music_developer_token,
            "music_user_token": apple_music_user_token,
            "is_public": True,
            "is_collaborative": False,
            "existing_urls": apple_music_url
        }
    if soundcloud_token:
        patch_args["soundcloud"] = {
            "user_id": soundcloud_user_id,
            "token": soundcloud_token,
            "is_public": True,
            "is_collaborative": False,
            "existing_urls": soundcloud_url
        }

    if args is None:
        args = []

    cmd = convert_patch_to_command(patches[patchname])
    context = cmd.make_context(patchname, list(args))
    patch_args.update(context.forward(cmd))

    # Create the actual patch, finally
    patch = patches[patchname](patch_args)
    if patch.is_local():
        logger.info(
            "This is a local patch and should be invoked via the specific troi function, rather than the playlist function.")
        return None

    try:
        ret = patch.generate_playlist()
    except RuntimeError as err:
        logger.error(err)
        ret = 0

    user_feedback = patch.user_feedback()
    if len(user_feedback) > 0:
        logger.info("User feedback:")
        for feedback in user_feedback:
            logger.info(f"  * {feedback}")
        logger.info("")

    sys.exit(0 if ret else -1)


@cli.command(name="list")
def list_patches_cli():
    """List all available patches"""
    ret = list_patches()
    sys.exit(0 if ret else -1)


@cli.command(name="info")
@click.argument("patch", nargs=1)
def info_cmd(patch):
    """Get info for a given patch"""
    ret = patch_info(patch)
    sys.exit(0 if ret else -1)


@cli.command(name="resolve", context_settings=dict(ignore_unknown_options=True, ))
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-t', '--threshold', default=.80, help="Minimum match percentage for metadata matches. Must be 0.0 - 1.0")
@click.option('-u', '--upload-to-subsonic', required=False, is_flag=True, help="upload playlist via subsonic API")
@click.option('-i', '--subsonic-id', required=False, help="overwrite existing subsonic playlist with the given ID")
@click.option('-m', '--save-to-m3u', required=False, help="save to specified m3u playlist")
@click.option('-j', '--save-to-jspf', required=False, help="save to specified JSPF playlist")
@click.option('-y', '--dont-ask', required=False, is_flag=True, help="save playlist without asking user")
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
@click.argument('jspf_playlist')
def resolve(db_file, threshold, upload_to_subsonic, subsonic_id, save_to_m3u, save_to_jspf, dont_ask, quiet, jspf_playlist):
    """ Resolve a global JSPF playlist with MusicBrainz MBIDs to files in the local collection"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = SubsonicDatabase(db_file, config, quiet)
    db.open()
    lbrl = ListenBrainzRadioLocal(quiet)
    playlist = read_jspf_playlist(jspf_playlist)
    lbrl.resolve_playlist(threshold, playlist)
    output_playlist(db, playlist, upload_to_subsonic, save_to_m3u, save_to_jspf, dont_ask, subsonic_id)


@cli.command(name="lb-radio", context_settings=dict(ignore_unknown_options=True, ))
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-t', '--threshold', default=.80, help="Minimum match percentage for metadata matches. Must be 0.0 - 1.0")
@click.option('-u', '--upload-to-subsonic', required=False, is_flag=True, help="upload playlist via subsonic API")
@click.option('-i', '--subsonic-id', required=False, help="overwrite existing subsonic playlist with the given ID")
@click.option('-m', '--save-to-m3u', required=False, help="save to specified m3u playlist")
@click.option('-j', '--save-to-jspf', required=False, help="save to specified JSPF playlist")
@click.option('-y', '--dont-ask', required=False, is_flag=True, help="save playlist without asking user")
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
@click.argument('mode')
@click.argument('prompt')
def lb_radio(db_file, threshold, upload_to_subsonic, subsonic_id, save_to_m3u, save_to_jspf, dont_ask, quiet, mode, prompt):
    """Use LB Radio to create a playlist from a prompt, using a local music collection"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = SubsonicDatabase(db_file, config, quiet)
    db.open()
    r = ListenBrainzRadioLocal(quiet)
    playlist = r.generate(mode, prompt, threshold)
    try:
        _ = playlist.playlists[0].recordings[0]
    except (KeyError, IndexError, AttributeError):
        db.metadata_sanity_check(include_subsonic=upload_to_subsonic)
        return

    output_playlist(db, playlist, upload_to_subsonic, save_to_m3u, save_to_jspf, dont_ask, subsonic_id)


@cli.command("weekly-jams", context_settings=dict(ignore_unknown_options=True, ))
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-t', '--threshold', default=.80, help="Minimum match percentage for metadata matches. Must be 0.0 - 1.0")
@click.option('-u', '--upload-to-subsonic', required=False, is_flag=True, help="upload playlist via subsonic API")
@click.option('-i', '--subsonic-id', required=False, help="overwrite existing subsonic playlist with the given ID")
@click.option('-m', '--save-to-m3u', required=False, help="save to specified m3u playlist")
@click.option('-j', '--save-to-jspf', required=False, help="save to specified JSPF playlist")
@click.option('-y', '--dont-ask', required=False, is_flag=True, help="save playlist without asking user")
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
@click.argument('user_name')
def periodic_jams(db_file, threshold, upload_to_subsonic, subsonic_id, save_to_m3u, save_to_jspf, dont_ask, quiet, user_name):
    "Generate a weekly jams playlist for your local collection"
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = SubsonicDatabase(db_file, config, quiet)
    db.open()

    pj = PeriodicJamsLocal(user_name, threshold, quiet)
    playlist = pj.generate()
    try:
        _ = playlist.playlists[0].recordings[0]
    except (KeyError, IndexError, AttributeError):
        db.metadata_sanity_check(include_subsonic=upload_to_subsonic)
        return

    output_playlist(db, playlist, upload_to_subsonic, save_to_m3u, save_to_jspf, dont_ask, subsonic_id)


@cli.command(context_settings=dict(ignore_unknown_options=True, ))
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def test(args):
    """Run unit tests"""
    import pytest
    raise SystemExit(pytest.main(list(args)))


if __name__ == "__main__":
    cli()
    sys.exit(0)



from abc import abstractmethod
from time import sleep

import requests

from troi import Recording, Artist, ArtistCredit
from troi.service import Service
from troi.plist import plist


class RecordingSearchByTagService(Service):

    SLUG = "recording-search-by-tag"

    def __init__(self):
        super().__init__(self.SLUG)

    def search(self, tags, operator, pop_begin, pop_end, num_recordings):
        """
            Fetch the tag data from the LB API and return it as a dict.
        """

        data = {
            "operator": operator,
            "count": num_recordings,
            "pop_begin": pop_begin,
            "pop_end": pop_end,
            "tag": tags
        }
        while True:
            r = requests.get("https://api.listenbrainz.org/1/lb-radio/tags", params=data)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch recordings for tags. {r.text}")

            break


        return plist([ Recording(mbid=rec["recording_mbid"], musicbrainz={"popularity": rec["percent"]}) for rec in r.json() ])


class RecordingSearchByArtistService(Service):

    SLUG = "recording-search-by-artist"

    def __init__(self):
        super().__init__(self.SLUG)

    def search(self, mode, artist_mbid, pop_begin, pop_end, max_recordings_per_artist, max_similar_artists):
        """
            Given a seed artist mbid, find and select similar artists (via LB similar artists data).

            pop_begin: The lower bound on recording popularity
            pop_end: The upper bound on recording popularity
            max_recordings_per_artist: The number of recordings to collect for each artist.
            max_similar_artists: The maximum number of similar artists to select.
        """

        params = {
                "mode": mode,
                "max_similar_artists": max_similar_artists,
                "max_recordings_per_artist": max_recordings_per_artist,
                "pop_begin": pop_begin,
                "pop_end": pop_end
        }
        url = f"https://api.listenbrainz.org/1/lb-radio/artist/{artist_mbid}"

        while True:
            r = requests.get(url, params=params)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch lb_radio artists: {r.status_code} ({r.text})")

            break


        try:
            artists = r.json()
        except IndexError:
            return {}, []

        artist_recordings = {}
        msgs = []
        for artist_mbid in artists:
            recordings = plist()
            for recording in artists[artist_mbid]:
                artist_credit = ArtistCredit(artists=[Artist(mbid=recording["similar_artist_mbid"])],
                                             name=recording["similar_artist_name"])
                recordings.append(Recording(mbid=recording["recording_mbid"],
                                            artist_credit=artist_credit,
                                            musicbrainz={"total_listen_count": recording["total_listen_count"]}))

            # Below is a hack, since the endpoint seems to return one track too few
            if len(recordings) < max_recordings_per_artist - 1:
                msgs.append("Artist %s has only few top recordings in %s mode" % (recordings[0].artist_credit.name, mode))

            artist_recordings[artist_mbid] = recordings.random_item(pop_begin, pop_end, max_recordings_per_artist)

        return artist_recordings, msgs



from collections import defaultdict
import troi
from random import randint
from troi import Recording
from troi import TARGET_NUMBER_OF_RECORDINGS


class InterleaveRecordingsElement(troi.Element):
    """
        This element round-robins the various input sources into one list until all sources are all empty.
    """

    def __init__(self):
        troi.Element.__init__(self)

    def inputs(self):
        return [Recording]

    def outputs(self):
        return [Recording]

    def read(self, entities):

        recordings = []
        while True:
            empty = 0
            for entity in entities:
                try:
                    recordings.append(entity.pop(0))
                except IndexError:
                    empty += 1

            # Did we process all the recordings?
            if empty == len(entities):
                break

        return recordings


class WeighAndBlendRecordingsElement(troi.Element):
    """
        This element will weight all the given sources according to weights passed to __init__ and
        then combine all the input sources into one weighted output stream.

        A source that has a weight of 2 will be chosen 2 times more often than a source with weight 1.
    """

    def __init__(self, weights, max_num_recordings=TARGET_NUMBER_OF_RECORDINGS, max_artist_occurrence=None):
        troi.Element.__init__(self)
        self.weights = weights
        self.max_num_recordings = max_num_recordings
        self.max_artist_occurrence = max_artist_occurrence

    def inputs(self):
        return [Recording]

    def outputs(self):
        return [Recording]

    def read(self, entities):

        total_available = sum([len(e) for e in entities])

        # prepare the weights
        total = sum(self.weights)
        summed = []
        acc = 0
        for i in self.weights:
            acc += i
            summed.append(acc)

        # Ensure seed artists are the first tracks -- doing this for all recording elements work in this case.
        recordings = []
        for element in entities:
            try:
                recordings.append(element.pop(0))
            except IndexError:
                pass

        # This still allows sequential tracks to be from the same artists. I'll wait for feedback to see if this
        # is a problem.
        artist_counts = defaultdict(int)
        dedup_set = set()
        while True:
            r = randint(0, total)
            for i, s in enumerate(summed):
                if r < s:
                    while True:
                        if len(entities[i]) > 0:
                            rec = entities[i].pop(0)
                            if rec.mbid in dedup_set:
                                total_available -= 1
                                continue
                            if self.max_artist_occurrence is not None and \
                                    artist_counts[rec.artist_credit.artist_credit_id] == self.max_artist_occurrence:
                                total_available -= 1
                                continue

                            recordings.append(rec)
                            dedup_set.add(rec.mbid)
                            artist_counts[rec.artist_credit.artist_credit_id] += 1
                        break

            if len(recordings) >= self.max_num_recordings or len(recordings) == total_available:
                break

        return recordings



import logging
import os

from troi.plist import plist
from troi import Recording as TroiRecording
from troi.content_resolver.model.recording import FileIdType

logger = logging.getLogger(__name__)


def ask_yes_no_question(prompt):

    while True:
        resp = input(prompt)
        resp = resp.strip()
        if resp == "":
            resp = 'y'

        if resp == 'y':
            return True
        elif resp == 'n':
            return False
        else:
            logging.info("eh? try again.")


def select_recordings_on_popularity(recordings, pop_begin, pop_end, num_recordings):
    """
       Given dicts of recording data, select up to num_recordings recordings randomly
       from the recordings that ideally lie in popularity between pop_begin and pop_end.

       If too little data is found in the percent range, select recordings that are the closest
       to the disired range.
    """

    matching_recordings = []
    over_recordings = []
    under_recordings = []
    for rec in recordings:
        if rec["popularity"] >= pop_begin:
            if rec["popularity"] < pop_end:
                matching_recordings.append(rec)
            else:
                over_recordings.append(rec)
        else:
            under_recordings.append(rec)

    # If we have enough recordings, skip the extending part
    if len(matching_recordings) < num_recordings:
        # We don't have enough recordings, see if we can pick the ones outside
        # of our desired range in a best effort to make a playlist.
        # Keep adding the best matches until we (hopefully) get our desired number of recordings
        while len(matching_recordings) < num_recordings:
            if under_recordings:
                under_diff = pop_begin - under_recordings[-1]["popularity"]
            else:
                under_diff = None

            if over_recordings:
                over_diff = over_recordings[-1]["popularity"] - pop_end
            else:
                over_diff = None

            if over_diff == None and under_diff == None:
                break

            if over_diff is not None and under_diff is not None and under_diff < over_diff:
                matching_recordings.insert(0, under_recordings.pop(-1))
            else:
                if under_diff is not None:
                    matching_recordings.insert(len(matching_recordings), under_recordings.pop(-1))
                else:
                    matching_recordings.insert(len(matching_recordings), over_recordings.pop(0))

    # Convert results into recordings
    results = plist()
    for rec in matching_recordings:
        r = TroiRecording(mbid=rec["recording_mbid"])
        if rec["file_id_type"] == FileIdType.SUBSONIC_ID:
            r.musicbrainz = {"subsonic_id": rec["file_id"]}
        if rec["file_id_type"] == FileIdType.FILE_PATH:
            r.musicbrainz = {"filename": rec["file_id"]}

        results.append(r)

    return results


class bcolors:
    """ Basic ASCII color codes """

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def existing_dirs(paths):
    """Yield absolute paths for all existing directories in the iterable passed"""
    for path in paths:
        abspath = os.path.abspath(path)
        if os.path.isdir(abspath):
            yield abspath



import unittest

import requests_mock

from troi.patch import Patch
from troi import Artist, ArtistCredit, Recording, Playlist
import troi.filters
from troi.musicbrainz.recording import RecordingListElement
from troi.listenbrainz.feedback import ListensFeedbackLookup


class DummyPatch:

    def __init__(self):
        self.local_storage = {}

    def set_patch_object(self, obj):
        pass


class TestArtistCreditFilterElement(unittest.TestCase):

    def test_artist_credit_filter_include(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', artist_credit=ArtistCredit(artist_credit_id=65)),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', artist_credit=ArtistCredit(artist_credit_id=197))
        ]
        alist = [197]

        e = troi.filters.ArtistCreditFilterElement(alist, include=True)
        flist = e.read([rlist])
        assert len(flist) == 1
        assert flist[0].artist_credit.artist_credit_id == 197
        assert flist[0].mbid == "73a9d0db-0ec7-490e-9a85-0525a5ccef8e"

        alist = [1]
        e = troi.filters.ArtistCreditFilterElement(alist, include=True)
        flist = e.read([rlist])
        assert len(flist) == 0

    def test_artist_credit_filter_exclude(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', artist_credit=ArtistCredit(artist_credit_id=65)),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', artist_credit=ArtistCredit(artist_credit_id=197))
        ]
        alist = [197]

        e = troi.filters.ArtistCreditFilterElement(alist)
        flist = e.read([rlist])
        assert len(flist) == 1
        assert flist[0].artist_credit.artist_credit_id == 65
        assert flist[0].mbid == "8756f690-18ca-488d-a456-680fdaf234bd"

        alist = [1]
        e = troi.filters.ArtistCreditFilterElement(alist)
        flist = e.read([rlist])
        assert len(flist) == 2


class TestArtistCreditLimiterElement(unittest.TestCase):

    def test_artist_credit_limiter_higher_ranked(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', artist_credit=ArtistCredit(artist_credit_id=65), ranking=1.0),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1', artist_credit=ArtistCredit(artist_credit_id=65), ranking=.5),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', artist_credit=ArtistCredit(artist_credit_id=197), ranking=.1)
        ]


        e = troi.filters.ArtistCreditLimiterElement(1)
        flist = e.read([rlist])
        assert len(flist) == 2
        assert flist[0].artist_credit.artist_credit_id == 65
        assert flist[0].mbid == "8756f690-18ca-488d-a456-680fdaf234bd"
        assert flist[1].artist_credit.artist_credit_id == 197
        assert flist[1].mbid == "73a9d0db-0ec7-490e-9a85-0525a5ccef8e"

        e = troi.filters.ArtistCreditLimiterElement(1, exclude_lower_ranked=False)
        flist = e.read([rlist])
        assert len(flist) == 2
        assert flist[0].artist_credit.artist_credit_id == 65
        assert flist[0].mbid == "139654ae-2c02-4e0f-aee0-c47da6e59ff1"
        assert flist[1].artist_credit.artist_credit_id == 197
        assert flist[1].mbid == "73a9d0db-0ec7-490e-9a85-0525a5ccef8e"

    def test_artist_credit_limiter_playlist(self):
        p = Playlist("test playlist")
        p.recordings = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', artist_credit=ArtistCredit(artist_credit_id=65), ranking=1.0),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1', artist_credit=ArtistCredit(artist_credit_id=65), ranking=.5),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', artist_credit=ArtistCredit(artist_credit_id=197), ranking=.1)
        ]

        p2 = Playlist("test playlist 2")
        p2.recordings = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', artist_credit=ArtistCredit(artist_credit_id=48), ranking=1.0),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', artist_credit=ArtistCredit(artist_credit_id=197), ranking=.1),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1', artist_credit=ArtistCredit(artist_credit_id=48), ranking=.5)
        ]

        e = troi.filters.ArtistCreditLimiterElement(1)
        plist = e.read([[p, p2]])
        assert len(plist) == 2
        assert len(plist[0].recordings) == 2
        assert plist[0].recordings[0].artist_credit.artist_credit_id == 65
        assert plist[0].recordings[0].mbid == "8756f690-18ca-488d-a456-680fdaf234bd"
        assert plist[0].recordings[1].artist_credit.artist_credit_id == 197
        assert plist[0].recordings[1].mbid == "73a9d0db-0ec7-490e-9a85-0525a5ccef8e"
        assert len(plist[1].recordings) == 2
        assert plist[1].recordings[0].artist_credit.artist_credit_id == 48
        assert plist[1].recordings[0].mbid == "8756f690-18ca-488d-a456-680fdaf234bd"
        assert plist[1].recordings[1].artist_credit.artist_credit_id == 197
        assert plist[1].recordings[1].mbid == "73a9d0db-0ec7-490e-9a85-0525a5ccef8e"


class TestDuplicateRecordingMBIDFilterElement(unittest.TestCase):

    def test_duplicate_recording_filter_element(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e'),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1')
        ]
        e = troi.filters.DuplicateRecordingMBIDFilterElement()
        flist = e.read([rlist])
        assert len(flist) == 3

        assert flist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[1].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'
        assert flist[2].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'


class TestConsecutiveRecordingFilterElement(unittest.TestCase):

    def test_consecutive_recording_filter_element(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1'),
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e'),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e'),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1')
        ]
        e = troi.filters.ConsecutiveRecordingFilterElement()
        flist = e.read([rlist])
        assert len(flist) == 5

        assert flist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[1].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'
        assert flist[2].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[3].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'
        assert flist[4].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'


class TestYearRangeFilterElement(unittest.TestCase):

    def test_year_range_filter_element(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', year=1980),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', year=1990),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1', year=2000)
        ]
        e = troi.filters.YearRangeFilterElement(1985, 1995)
        flist = e.read([rlist])
        assert len(flist) == 1

        assert flist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'

        e = troi.filters.YearRangeFilterElement(1985, 1995, inverse=True)
        flist = e.read([rlist])
        assert len(flist) == 2

        assert flist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[1].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'

        e = troi.filters.YearRangeFilterElement(1985)
        flist = e.read([rlist])
        assert len(flist) == 2

        assert flist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'
        assert flist[1].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'

        e = troi.filters.YearRangeFilterElement(1985, inverse=True)
        flist = e.read([rlist])
        assert len(flist) == 1

        assert flist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'


class TestHatedRecordingsFilterElement(unittest.TestCase):

    @requests_mock.Mocker()
    def test_hated_recordings_filter(self, mock_requests):
        mock_requests.get("https://api.listenbrainz.org/1/feedback/user/lucifer/get-feedback-for-recordings",
                          json={
                              "feedback": [{
                                  "created": None,
                                  "recording_mbid": "53969964-673a-4407-9396-3087be9245f6",
                                  "recording_msid": None,
                                  "score": 1,
                                  "track_metadata": None,
                                  "user_id": "lucifer"
                              }, {
                                  "created": None,
                                  "recording_mbid": "70c75409-4224-4bab-836a-eb5f3a9c31d3",
                                  "recording_msid": None,
                                  "score": -1,
                                  "track_metadata": None,
                                  "user_id": "lucifer"
                              }]
                          })

        recordings = RecordingListElement([
            Recording(mbid="53969964-673a-4407-9396-3087be9245f6"),
            Recording(mbid="70c75409-4224-4bab-836a-eb5f3a9c31d3"),
            Recording(mbid="8e7a9ff8-c31d-4ac0-a01d-20a7fcc28c8f")
        ])
        feedback_lookup = ListensFeedbackLookup(user_name="lucifer")
        feedback_lookup.set_sources(recordings)
        filter_element = troi.filters.HatedRecordingsFilterElement()
        filter_element.set_sources(feedback_lookup)

        received = filter_element.generate(quiet=True)
        expected = [
            Recording(mbid="53969964-673a-4407-9396-3087be9245f6", listenbrainz={"score": 1}),
            Recording(mbid="8e7a9ff8-c31d-4ac0-a01d-20a7fcc28c8f", listenbrainz={"score": 0})
        ]
        self.assertEqual(expected[0].mbid, received[0].mbid)
        self.assertEqual(expected[0].listenbrainz, received[0].listenbrainz)
        self.assertEqual(expected[1].mbid, received[1].mbid)
        self.assertEqual(expected[1].listenbrainz, received[1].listenbrainz)


class TestNeverListenedFilterElement(unittest.TestCase):

    def test_never_listened_element(self):
        rlist = [
            Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd', listenbrainz={"latest_listened_at": 342134214}),
            Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e', listenbrainz={"latest_listened_at": None}),
            Recording(mbid='c88b3490-35a0-460d-b3bc-bc50c8855d00'),
            Recording(mbid='139654ae-2c02-4e0f-aee0-c47da6e59ff1', listenbrainz={"latest_listened_at": 345345345})
        ]
        p = DummyPatch()
        e = troi.filters.NeverListenedFilterElement()
        e.set_patch_object(p)
        flist = e.read([rlist])
        assert len(flist) == 2

        assert flist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert flist[1].mbid == '139654ae-2c02-4e0f-aee0-c47da6e59ff1'

        e = troi.filters.NeverListenedFilterElement(remove_unlistened=False)
        e.set_patch_object(p)
        flist = e.read([rlist])
        assert len(flist) == 2

        assert flist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'
        assert flist[1].mbid == 'c88b3490-35a0-460d-b3bc-bc50c8855d00'



import troi.external.gpt
import troi.musicbrainz.mbid_mapping
from troi import Playlist
from troi.patch import Patch
from troi.playlist import PlaylistMakerElement


class AiJamsPatch(Patch):
    """ Generate a playlist using AI from the given prompt. """

    @staticmethod
    def inputs():
        """
        Generate a playlist using AI

        \b
        API_KEY is the OpenAI api key.
        PROMPT is the description of the playlist to generate.
        """
        return [
            {"type": "argument", "args": ["api_key"]},
            {"type": "argument", "args": ["prompt"]}
        ]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "ai-jams"

    @staticmethod
    def description():
        return "Generate a playlist using AI from the given prompt."

    def create(self, inputs):
        api_key = inputs['api_key']
        prompt = inputs['prompt'].strip()

        ai_recordings_lookup = troi.external.gpt.GPTRecordingElement(api_key, prompt)

        recs_lookup = troi.musicbrainz.mbid_mapping.MBIDMappingLookupElement(remove_unmatched=True)
        recs_lookup.set_sources(ai_recordings_lookup)

        pl_maker = PlaylistMakerElement(
            patch_slug=self.slug(),
            max_num_recordings=50,
            max_artist_occurrence=2,
            shuffle=False
        )
        pl_maker.set_sources(recs_lookup)

        return pl_maker



import unittest
import unittest.mock

import troi
import troi.musicbrainz.related_artist_credits

return_json = [
    {
        "artist_credit_id": 65,
        "artist_credit_name": "Portishead",
        "count": 64,
        "related_artist_credit_id": 20660,
        "related_artist_credit_name": "Oasis"
    },
    {
        "artist_credit_id": 963,
        "artist_credit_name": "Morcheeba",
        "count": 120,
        "related_artist_credit_id": 359,
        "related_artist_credit_name": "Moby"
    }
]

class TestArtistCreditNameLookup(unittest.TestCase):

    @unittest.mock.patch('requests.get')
    def test_read(self, req):

        mock = unittest.mock.MagicMock()
        mock.status_code = 200
        mock.text = return_json
        req.return_value = mock
        e = troi.musicbrainz.related_artist_credits.RelatedArtistCreditsElement()

        inputs = [ troi.ArtistCredit(artist_credit_id=65), 
                   troi.ArtistCredit(artist_credit_id=963) ]
        entities = e.read([inputs])
        req.assert_called_with(e.SERVER_URL, params={'[artist_credit_id]': '65,963', 'threshold': 0})

        assert len(entities) == 2
        assert entities[0].artist_credit_id == 65
        assert entities[1].artist_credit_id == 963
        assert len(entities[0].musicbrainz['related_artist_credit_ids']) == 1

        assert entities[0].musicbrainz['related_artist_credit_ids'][0]['artist_credit_id'] == 65
        assert entities[0].musicbrainz['related_artist_credit_ids'][0]['artist_credit_name'] == 'Portishead'
        assert entities[0].musicbrainz['related_artist_credit_ids'][0]['related_artist_credit_id'] == 20660
        assert entities[0].musicbrainz['related_artist_credit_ids'][0]['related_artist_credit_name'] == 'Oasis'

        assert len(entities[1].musicbrainz['related_artist_credit_ids']) == 1
        assert entities[1].musicbrainz['related_artist_credit_ids'][0]['artist_credit_id'] == 963
        assert entities[1].musicbrainz['related_artist_credit_ids'][0]['artist_credit_name'] == 'Morcheeba'
        assert entities[1].musicbrainz['related_artist_credit_ids'][0]['related_artist_credit_id'] == 359
        assert entities[1].musicbrainz['related_artist_credit_ids'][0]['related_artist_credit_name'] == 'Moby'



import json

from troi import Playlist
from troi.patch import Patch
from troi.playlist import PlaylistFromJSPFElement


class TransferPlaylistPatch(Patch):

    @staticmethod
    def inputs():
        """
        A dummy patch that retrieves an existing playlist from ListenBrainz or raw JSPF for use in Troi.

        \b
        MBID is the playlist mbid to save. Specify this OR JSPF.
        JSPF is the actual JSPF playlist to transfer. Specify this OR MBID.
        READ_ONLY_TOKEN is the listenbrainz auth token to retrieve the playlist if its private. If not specified,
        fallback to TOKEN. Both arguments take the same value but specifying TOKEN may also upload the playlist
        to LB again which is many times not desirable.
        """
        return [
            {"type": "argument", "args": ["mbid"], "kwargs": {"required": False}},
            {"type": "argument", "args": ["jspf"], "kwargs": {"required": False}},
            {"type": "argument", "args": ["read_only_token"], "kwargs": {"required": False}}
        ]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "transfer-playlist"

    @staticmethod
    def description():
        return "Retrieve a playlist from the ListenBrainz"

    def create(self, inputs):

        mbid = inputs["mbid"]
        jspf = inputs["jspf"]

        if mbid == "":
            mbid = None
        if jspf == "":
            jspf = None

        if mbid is not None and jspf is not None:
            raise RuntimeError("Cannot pass both playlist mbid and jspf to TransferPlaylistPatch.")

        if mbid is None and jspf is None:
            raise RuntimeError("Must pass either playlist mbid or jspf to TransferPlaylistPatch.")

        if isinstance(jspf, str):
            jspf = json.loads(jspf)

        token = inputs.get("read_only_token") or inputs.get("token")
        return PlaylistFromJSPFElement(playlist_mbid=mbid, jspf=jspf, token=token)



import unittest

from troi import Artist, ArtistCredit, Release, Recording, Playlist


class TestEntities(unittest.TestCase):

    def test_artist(self):

        a = Artist()
        assert a.mbid is None
        assert a.name is None

        a = Artist("Portishead", mbid="8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", ranking=1.0)
        assert a.name == "Portishead"
        assert a.mbid == "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3"
        assert a.ranking == 1.0

        a = Artist(listenbrainz={ 1 : 2}, musicbrainz={ 3 : 4}, acousticbrainz={ 5 : 6})
        assert a.lb[1] == 2
        assert a.mb[3] == 4
        assert a.ab[5] == 6
        assert a.name is None
        assert a.mbid is None

    def test_artist_credit(self):

        a = ArtistCredit()
        assert a.artists == []
        assert a.name is None
        assert a.artist_credit_id is None
        with self.assertRaises(TypeError):
            a = Artist(artists="not a list")

        a = ArtistCredit("Portishead", artists=[Artist(name="Portishead", mbid="8fe3f1d4-b6d5-4726-89ad-926e3420b9e3")], ranking=1.0)
        assert a.name == "Portishead"
        assert len(a.artists) == 1
        assert a.ranking == 1.0

        a = ArtistCredit("Portishead", artist_credit_id=65)
        assert a.artist_credit_id == 65

        a = Artist(listenbrainz={ 1 : 2}, musicbrainz={ 3 : 4}, acousticbrainz={ 5 : 6})
        assert a.lb[1] == 2
        assert a.mb[3] == 4
        assert a.ab[5] == 6
        assert a.name is None
        assert a.mbid is None

    def test_release(self):

        r = Release()
        assert r.mbid is None
        assert r.name is None

        r = Release("Dummy", "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", ranking=.5)
        assert r.name == "Dummy"
        assert r.mbid == "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3"
        assert r.ranking == .5

        r = Release(listenbrainz={ 1 : 2}, musicbrainz={ 3 : 4}, acousticbrainz={ 5 : 6})
        assert r.lb[1] == 2
        assert r.mb[3] == 4
        assert r.ab[5] == 6
        assert r.name is None
        assert r.mbid is None

    def test_recording(self):

        r = Recording()
        assert r.mbid is None
        assert r.msid is None
        assert r.name is None

        r = Recording("Strangers", "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", 
                      "97b01626-65fc-4c32-b30c-c4d7eab1b339", ranking=.1, year=1995)
        assert r.name == "Strangers"
        assert r.mbid == "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3"
        assert r.msid == "97b01626-65fc-4c32-b30c-c4d7eab1b339"
        assert r.ranking == .1
        assert r.year == 1995

        r = Recording(listenbrainz={ 1 : 2}, musicbrainz={ 3 : 4}, acousticbrainz={ 5 : 6})
        assert r.lb[1] == 2
        assert r.mb[3] == 4
        assert r.ab[5] == 6
        assert r.name is None
        assert r.mbid is None
        assert r.msid is None


    def test_playlist(self):

        r = Playlist()
        assert r.mbid is None
        assert r.name is None
        assert r.filename is None
        assert r.recordings == None

        p = Playlist("cooking playlist", "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", "mk_noms.jspf")
        assert p.name == "cooking playlist"
        assert p.mbid == "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3"
        assert p.filename == "mk_noms.jspf"
        assert p.recordings == None

        r = Recording("Strangers", "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", "97b01626-65fc-4c32-b30c-c4d7eab1b339")
        p = Playlist("eating playlist", "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3", "omnomnom.jspf", [r])
        assert p.name == "eating playlist"
        assert p.mbid == "8fe3f1d4-b6d5-4726-89ad-926e3420b9e3"
        assert p.filename == "omnomnom.jspf"
        assert p.recordings == [r]



from troi import Element, User


class UserListElement(Element):
    '''
        This element is used to pass a provided list of users into the pipeline.

        :param user_list: A list of user_names
    '''

    def __init__(self, user_list):
        super().__init__()
        self.user_list = user_list

    @staticmethod
    def inputs():
        return []

    @staticmethod
    def outputs():
        return [User]

    def read(self, inputs):
        return [ User(user_name=user_name) for user_name in self.user_list ]



import unittest

from troi import Artist, ArtistCredit, Release, Recording
import troi.operations 


class TestOperations(unittest.TestCase):

    def test_is_homogeneous(self):
        alist = [ ]
        assert troi.operations.is_homogeneous(alist) == True

        alist = [ Artist(), Artist() ]
        assert troi.operations.is_homogeneous(alist) == True

        alist = [ Artist(), Release() ]
        assert troi.operations.is_homogeneous(alist) == False

    def test_unique(self):
        # Test artists
        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        e = troi.operations.UniqueElement('mbid')
        u_alist = e.read([alist])
        assert len(u_alist) == 2

        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        e = troi.operations.UniqueElement('mbid')
        u_alist = e.read([alist])
        assert len(u_alist) == 1
        assert u_alist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'

        alist = [ ArtistCredit(artist_credit_id=65),
                  ArtistCredit(artist_credit_id=65) ]
        e = troi.operations.UniqueElement('artist_credit_id')
        u_alist = e.read([alist])
        assert len(u_alist) == 1
        assert u_alist[0].artist_credit_id == 65

        # Test recording (not testing release since rel and rec use same code)
        rlist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        e = troi.operations.UniqueElement('foo')
        with self.assertRaises(ValueError):
            u_rlist = e.read([rlist])

        e = troi.operations.UniqueElement('mbid')
        u_rlist = e.read([rlist])
        assert len(u_rlist) == 2

        rlist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        u_rlist = e.read([rlist])
        assert len(u_rlist) == 1
        assert u_rlist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'

    def test_ensure_conformity(self):
        # Test artists
        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Artist(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        assert troi.operations._ensure_conformity(alist, blist) == True

        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Release(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        with self.assertRaises(TypeError):
            troi.operations._ensure_conformity(alist, blist)

        # Test recording (not testing release since rel and rec use same code)
        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Recording(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        assert troi.operations._ensure_conformity(alist, blist) == True

        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Release(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        with self.assertRaises(TypeError):
            troi.operations._ensure_conformity(alist, blist)

    def test_union(self):
        # Test artists
        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        blist = [ Artist(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        e = troi.operations.UnionElement()
        ulist = e.read([alist, blist])
        assert len(ulist) == 2
        assert ulist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert ulist[1].mbid == 'a1c35a51-d102-4ce7-aefb-79a361e843b6'

        # Test recording (not testing release since rel and rec use same code)
        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        blist = [ Recording(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        ulist = e.read([alist, blist])
        assert len(ulist) == 2
        assert ulist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert ulist[1].mbid == 'a1c35a51-d102-4ce7-aefb-79a361e843b6'

    def test_intersection(self):
        # Test artists
        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        e = troi.operations.IntersectionElement('meh')
        with self.assertRaises(ValueError):
            ilist = e.read([alist, blist])

        e = troi.operations.IntersectionElement('mbid')
        ilist = e.read([alist, blist])
        assert len(ilist) == 1
        assert ilist[0].mbid == blist[0].mbid

        alist = [ Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        ilist = e.read([alist, blist])
        assert len(ilist) == 0

        # Test recording (not testing release since rel and rec use same code)
        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        e = troi.operations.IntersectionElement('mbids')
        with self.assertRaises(ValueError):
            ilist = e.read([alist, blist])

        e = troi.operations.IntersectionElement('mbid')
        ilist = e.read([alist, blist])
        assert len(ilist) == 1
        assert ilist[0].mbid == blist[0].mbid

        alist = [ Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        ilist = e.read([alist, blist])
        assert len(ilist) == 0

    def test_difference(self):
        # Test artists
        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        e = troi.operations.DifferenceElement('mbids')
        with self.assertRaises(ValueError):
            ilist = e.read([alist, blist])

        e = troi.operations.DifferenceElement('mbid')
        ilist = e.read([alist, blist])
        assert len(ilist) == 1
        assert ilist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'

        alist = [ Artist(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Artist(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Artist(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        dlist = e.read([alist, blist])
        assert len(dlist) == 2
        assert dlist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert dlist[1].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'

        # Test recording (not testing release since rel and rec use same code)
        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd') ]
        e = troi.operations.DifferenceElement('mbids')
        with self.assertRaises(ValueError):
            ilist = e.read([alist, blist])

        e = troi.operations.DifferenceElement('mbid')
        ilist = e.read([alist, blist])
        assert len(ilist) == 1
        assert ilist[0].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'

        alist = [ Recording(mbid='8756f690-18ca-488d-a456-680fdaf234bd'), 
                  Recording(mbid='73a9d0db-0ec7-490e-9a85-0525a5ccef8e') ]
        blist = [ Recording(mbid='a1c35a51-d102-4ce7-aefb-79a361e843b6') ]
        dlist = e.read([alist, blist])
        assert len(dlist) == 2
        assert dlist[0].mbid == '8756f690-18ca-488d-a456-680fdaf234bd'
        assert dlist[1].mbid == '73a9d0db-0ec7-490e-9a85-0525a5ccef8e'



import unittest

from troi.utils import discover_patches
from troi.patch import Patch


class TestPatches(unittest.TestCase):

    def test_discover_patches(self):
        patches = discover_patches()

        assert "periodic-jams" in patches
        assert "playlist-from-mbids" in patches
        assert "recs-to-playlist" in patches
        assert "transfer-playlist" in patches

        for p in patches:
            assert issubclass(patches[p], Patch)



from troi import Element, Recording


class RecordingListElement(Element):
    """
        This element is used to pass a provided list of Recordings into the pipeline.

        :param recordings: The recordings to return from this Element.
    """

    def __init__(self, recordings):
        super().__init__()
        self.recordings = recordings

    @staticmethod
    def inputs():
        return []

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        return self.recordings



import datetime

from troi import PipelineError, Recording, DEVELOPMENT_SERVER_URL 
import troi.tools.area_lookup
import troi.musicbrainz.recording_lookup
import troi.patch
import troi.filters
import troi.playlist
from troi.listenbrainz.dataset_fetcher import DataSetFetcherElement


class AreaRandomRecordingsPatch(troi.patch.Patch):

    SERVER_URL = DEVELOPMENT_SERVER_URL + "/area-random-recordings/json"

    def __init__(self, args):
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Generate a list of random recordings from a given area.

        \b
        AREA is a MusicBrainz area from which to choose tracks.
        START_YEAR is the start year.
        END_YEAR is the end year.
        """
        return [
            {"type": "argument", "args": ["area"]},
            {"type": "argument", "args": ["start_year"], "kwargs": {"type": int}},
            {"type": "argument", "args": ["end_year"], "kwargs": {"type": int}}
        ]

    @staticmethod
    def outputs():
        return [Recording]

    @staticmethod
    def slug():
        return "area-random-recordings"

    @staticmethod
    def description():
        return "Generate a list of random recordings from a given area."

    def create(self, inputs):
        area_name = inputs['area']
        start_year = inputs['start_year']
        end_year = inputs['end_year']

        area_id = troi.tools.area_lookup.area_lookup(area_name)

        if not start_year or start_year < 1800 or start_year > datetime.datetime.today().year:
            raise PipelineError("start_year must be given and be an integer between 1800 and the current year.")

        if not end_year or end_year < 1800 or end_year > datetime.datetime.today().year:
            raise PipelineError("end_year must be given and be an integer between 1800 and the current year.")

        if end_year < start_year:
            raise PipelineError("end_year must be equal to or greater than start_year.")

        recs = DataSetFetcherElement(server_url=self.SERVER_URL,
                                     json_post_data=[{ 'start_year': start_year,
                                                       'end_year': end_year,
                                                       'area_mbid': area_id }])

        name = "Random recordings from %s between %d and %d." % (area_name, start_year, end_year)
        pl_maker = troi.playlist.PlaylistMakerElement(name=name, desc=name, max_artist_occurrence=2)
        pl_maker.set_sources(recs)

        return pl_maker



#!/usr/bin/env python3
import logging
import sys

import click

from troi.logging_utils import set_log_level
from troi.content_resolver.database import Database
from troi.content_resolver.subsonic import SubsonicDatabase
from troi.content_resolver.metadata_lookup import MetadataLookup
from troi.content_resolver.utils import ask_yes_no_question
from troi.content_resolver.top_tags import TopTags
from troi.content_resolver.duplicates import FindDuplicates
from troi.content_resolver.playlist import write_m3u_playlist, write_jspf_playlist
from troi.content_resolver.unresolved_recording import UnresolvedRecordingTracker

logger = logging.getLogger(__name__)

# TODO: Soon we will need a better configuration file, so we can get rid of this hack
try:
    sys.path.insert(1, ".")
    import config
except ImportError as err:
    config = None

DEFAULT_CHUNKSIZE = 100


def output_playlist(db, playlist, upload_to_subsonic, save_to_m3u, save_to_jspf, dont_ask, subsonic_id=None):
    try:
        recording = playlist.playlists[0].recordings[0]
    except (KeyError, IndexError):
        logger.error("Cannot save empty playlist.")
        return

    if upload_to_subsonic and config:
        if recording and config.SUBSONIC_HOST:
            try:
                _ = recording.musicbrainz["subsonic_id"]
            except KeyError:
                logger.info("Playlist does not appear to contain subsonic ids. Can't upload to subsonic.")
                return

            if dont_ask or ask_yes_no_question("Upload via subsonic? (Y/n)"):
                logger.info("uploading playlist")
                db.upload_playlist(playlist, subsonic_id)
            return

    if save_to_m3u or save_to_jspf:
        try:
            _ = recording.musicbrainz["filename"]
        except KeyError:
            logger.error("Playlist does not appear to contain file paths. Can't write a local playlist.")
            return

    if save_to_m3u:
        if dont_ask or ask_yes_no_question(f"Save to '{save_to_m3u}'? (Y/n)"):
            logger.info("saving playlist")
            write_m3u_playlist(save_to_m3u, playlist)
        return

    if save_to_jspf:
        if dont_ask or ask_yes_no_question(f"Save to '{save_to_jspf}'? (Y/n)"):
            logger.info("saving playlist")
            write_jspf_playlist(save_to_jspf, playlist)
        return

    logger.info("Playlist displayed, but not saved. Use -j, -m or -u options to save/upload playlists.")


def db_file_check(db_file):
    """ Check the db_file argument and give useful user feedback. """

    if not db_file:
        if not config:
            logger.info("Database file not specified with -d (--db_file) argument. Consider adding it to config.py for ease of use.")
            sys.exit(-1)

        if not config.DATABASE_FILE:
            logger.info("config.py found, but DATABASE_FILE is empty. Please add it or use -d option to specify it.")
            sys.exit(-1)

        return config.DATABASE_FILE
    else:
        return db_file


def music_directories_from_config():
    """ Returns list of music directories if any in config file. """

    try:
        return list(set(config.MUSIC_DIRECTORIES))
    except AttributeError:
        return []


@click.group()
def cli():
    pass


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('--quiet', '-q', 'quiet', help="Do no print out anything", required=False, is_flag=True)
def create(db_file, quiet):
    """Create a new database to track a music collection"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = Database(db_file, quiet)
    db.create()


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-c', '--chunksize', default=DEFAULT_CHUNKSIZE, help="Number of files to add/update at once")
@click.option("-f", "--force", required=False, is_flag=True, default=False, help="Force scanning, ignoring any cache")
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
@click.argument('music_dirs', nargs=-1, type=click.Path())
def scan(db_file, music_dirs, quiet, chunksize=DEFAULT_CHUNKSIZE, force=False):
    """Scan one or more directories and their subdirectories for music files to add to the collection.
       If no path is passed, check for MUSIC_DIRECTORIES in config instead.
    """
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = Database(db_file, quiet)
    db.open()
    if not music_dirs:
        music_dirs = music_directories_from_config()
    db.scan(music_dirs, chunksize=chunksize, force=force)

    # Remove any recordings from the unresolved recordings that may have just been added.
    urt = UnresolvedRecordingTracker()
    releases = urt.cleanup()


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option("-r", "--remove", help="Without this flag, no files are removed.", required=False, is_flag=True, default=True)
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
def cleanup(db_file, remove, quiet):
    """Perform a database cleanup. Check that files exist and if they don't remove from the index"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = Database(db_file, quiet)
    db.open()
    db.database_cleanup(remove)


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
def metadata(db_file, quiet):
    """Lookup metadata (popularity and tags) for recordings"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = Database(db_file, quiet)
    db.open()
    lookup = MetadataLookup(quiet)
    lookup.lookup()

    logger.info("\nThese top tags describe your collection:")
    tt = TopTags()
    tt.print_top_tags_tightly(100)


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-q', '--quiet', 'quiet', help="Do no print out anything", required=False, is_flag=True)
def subsonic(db_file, quiet):
    """Scan a remote subsonic music collection"""
    set_log_level(quiet)
    db_file = db_file_check(db_file)
    db = SubsonicDatabase(db_file, config, quiet)
    db.open()
    db.sync()


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.argument('count', required=False, default=250)
def top_tags(db_file, count):
    "Display the top most used tags in the music collection. Useful for writing LB Radio tag prompts"
    set_log_level(False)
    db_file = db_file_check(db_file)
    db = Database(db_file, False)
    db.open()
    tt = TopTags()
    tt.print_top_tags_tightly(count)


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
@click.option('-e',
              '--exclude-different-release',
              help="Exclude duplicates that appear on different releases",
              required=False,
              default=False,
              is_flag=True)
@click.option('-v', '--verbose', help="Display extra info about found files", required=False, default=False, is_flag=True)
def duplicates(db_file, exclude_different_release, verbose):
    "Print all the tracks in the DB that are duplicated as per recording_mbid"
    set_log_level(False)
    db_file = db_file_check(db_file)
    db = Database(db_file, False)
    db.open()
    fd = FindDuplicates(db)
    fd.print_duplicate_recordings(exclude_different_release, verbose)


@click.command()
@click.option("-d", "--db_file", help="Database file for the local collection", required=False, is_flag=False)
def unresolved(db_file):
    "Show the top unresolved releases"
    set_log_level(False)
    db_file = db_file_check(db_file)
    db = Database(db_file, False)
    db.open()
    urt = UnresolvedRecordingTracker()
    releases = urt.get_releases()
    urt.print_releases(releases)


cli.add_command(create)
cli.add_command(scan)
cli.add_command(cleanup)
cli.add_command(metadata)
cli.add_command(subsonic)
cli.add_command(top_tags)
cli.add_command(duplicates)
cli.add_command(unresolved)


def usage(command):
    with click.Context(command) as ctx:
        click.echo(command.get_help(ctx))


if __name__ == "__main__":
    cli()
    sys.exit(0)



import logging

from troi.content_resolver.model.database import db

logger = logging.getLogger(__name__)


class TopTags:
    '''
       Class to fetch top tags
    '''

    def get_top_tags(self, limit=50):
        """
        """

        query = """SELECT tag.name
                        , COUNT(tag.id) AS cnt
                     FROM tag
                     JOIN recording_tag
                       ON recording_tag.tag_id = tag.id
                     JOIN recording
                       ON recording_tag.recording_id = recording.id
                 GROUP BY tag.name
                 ORDER BY cnt DESC
                    LIMIT ?"""

        cursor = db.execute_sql(query, (limit,))

        top_tags = []
        for rec in cursor.fetchall():
            top_tags.append({"tag": rec[0], "count": rec[1]})

        return top_tags

    def print_top_tags(self, limit=50):

        top_tags = self.get_top_tags(limit)
        for tt in top_tags:
            logger.info("%-40s %d" % (tt["tag"], tt["count"]))
        logger.info("")

    def print_top_tags_tightly(self, limit=250):

        top_tags = self.get_top_tags(limit)

        logger.info("; ".join(["%s %s" % (tt["tag"], tt["count"]) for tt in top_tags]))



import json
import requests
from troi import Element, Recording, PipelineError
import liblistenbrainz
import liblistenbrainz.errors
import dateutil.parser

MAX_NUM_RECORDINGS_PER_REQUEST = 100


class UserRecordingRecommendationsElement(Element):
    '''
        Fetch recommended recordings for a user from ListenBrainz.

        :param user_name: The ListenBrainz user to fetch recs for.
        :param artist_type: The type of recs to fetch. Must be one of "top", "similar" or "raw".
        :param count: The number of recs to fetch. Defaults to 25.
    '''

    MAX_RECORDINGS_TO_FETCH = 2000

    def __init__(self, user_name, artist_type, count=25, offset=0, auth_token=None):
        super().__init__()
        self.client = liblistenbrainz.ListenBrainz()
        if auth_token is not None:
            self.client.set_auth_token(auth_token)
        self.user_name = user_name
        self.count = count
        self.offset = offset
        self.artist_type = artist_type
        self._last_updated = None

    def outputs(self):
        return [Recording]

    @property
    def last_updated(self):
        return self._last_updated

    def read(self, inputs = []):
        recording_list = []

        remaining = self.MAX_RECORDINGS_TO_FETCH if self.count < 0 else self.count
        while True:
            try:
                recordings = self.client.get_user_recommendation_recordings(self.user_name, 
                                                                            self.artist_type, 
                                                                            count=min(MAX_NUM_RECORDINGS_PER_REQUEST, remaining),
                                                                            offset=self.offset+len(recording_list))
            except (requests.exceptions.HTTPError,
                    liblistenbrainz.errors.ListenBrainzAPIException,
                    requests.exceptions.ConnectionError) as err:
                if not str(err):
                    err = "Does the user '%s' exist?" % self.user_name
                raise PipelineError("Cannot fetch recommeded tracks from ListenBrainz: " + str(err))

            if not recordings or not len(recordings['payload']['mbids']):
                break

            for r in recordings['payload']['mbids']:
                latest = r.get("latest_listened_at", None)
                if latest is not None:
                    latest = dateutil.parser.isoparse(latest)
                    latest = latest.replace(tzinfo=None)
                lb_metadata = { "model_id": recordings["payload"].get("model_id", None),
                                "model_url": recordings["payload"].get("model_url", None),
                                "latest_listened_at": latest }
                recording_list.append(Recording(mbid=r['recording_mbid'], ranking=r['score'], listenbrainz=lb_metadata))

            remaining -= len(recordings['payload']['mbids'])
            if remaining <= 0:
                break

        if recordings:
            self._last_updated = recordings['payload']['last_updated']

        return recording_list



import logging
from copy import copy

import troi
from troi import PipelineError, User
from troi.utils import discover_patches

logger = logging.getLogger(__name__)


class ForLoopElement(troi.Element):
    '''
        An element that receives items from a pipeline and for each item in the pipeline it instantiates a new patch and executes that patch. A normal use case might include taking a list of users and running the specified patch for each user.

        As of right now, only User objects can be processed with this element.

        :param patch_slug: the slug of the patch to run inside the for loop
        :param pipeline_args: the arguments passed to top-level pipeline, so that they apply to dynamically created pipelines inside the for loop
    '''

    def __init__(self, patch_slugs, patch_args):
        super().__init__()
        self.patch_slugs = patch_slugs
        self.patch_args = patch_args

    @staticmethod
    def inputs():
        return [User]

    @staticmethod
    def outputs():
        return []

    def read(self, inputs):

        patches = discover_patches()

        outputs = []
        for patch_slug in self.patch_slugs:
            if not patch_slug in patches:
                raise PipelineError("ForLoop: Cannot load patch '%s'" % patch_slug)

            patch = patches[patch_slug](False)
            for user in inputs[0]:
                args = copy(self.patch_args)
                args["user_name"] = user.user_name
                pipeline = patch.create(args)
                self.patch_args["created_for"] = user.user_name

                try:
                    logger.info("generate %s for %s" % (patch_slug, user.user_name))
                    playlist = troi.playlist.PlaylistElement()
                    playlist.set_sources(pipeline)
                    playlist.generate()

                    if self.patch_args["min_recordings"] is not None and \
                        len(playlist.playlists[0].recordings) < self.patch_args["min_recordings"]:
                        logger.info("Playlist does not have at least %d recordings, not submitting.\n" % self.patch_args["min_recordings"])
                        continue

                    if not self.patch_args["quiet"]:
                        playlist.print()
                    playlist.add_metadata({"algorithm_metadata": {"source_patch": patch_slug}})
                    if self.patch_args["upload"]:
                        if not self.patch_args["token"] or self.patch_args["token"] == "":
                            raise PipelineError("In order to upload a playlist an auth token must be provided. Use --token")

                        try:
                            if self.patch_args["created_for"] and self.patch_args["created_for"] != "":
                                playlist.submit(self.patch_args["token"], self.patch_args["created_for"])
                            else:
                                playlist.submit(self.patch_args["token"])
                        except troi.PipelineError as err:
                            logger.error("Failed to submit playlist: %s, continuing..." % err)
                            continue

                    outputs.append(playlist.playlists[0])
                    logger.info("")
                except troi.PipelineError as err:
                    logger.info("Failed to generate playlist: %s" % err)
                    raise

        # Return None if you want to stop processing this pipeline
        return outputs



import logging
from collections import defaultdict
import re
import requests
import spotipy
from more_itertools import chunked
from spotipy import SpotifyException

logger = logging.getLogger(__name__)

APPLE_MUSIC_URL = f"https://api.music.apple.com/"
SPOTIFY_IDS_LOOKUP_URL = "https://labs.api.listenbrainz.org/spotify-id-from-mbid/json"
CLEAN_HTML_RE = re.compile('<.*?>')

def lookup_spotify_ids(recordings):
    """ Given a list of Recording elements, try to find spotify track ids from labs api spotify lookup using mbids
    and add those to the recordings. """
    response = requests.post(
        SPOTIFY_IDS_LOOKUP_URL,
        json=[{"recording_mbid": recording.mbid} for recording in recordings]
    )
    response.raise_for_status()
    spotify_data = response.json()
    mbid_spotify_ids_index = {}
    spotify_id_mbid_index = {}
    for recording, lookup in zip(recordings, spotify_data):
        if len(lookup["spotify_track_ids"]) > 0:
            recording.spotify_id = lookup["spotify_track_ids"][0]
            mbid_spotify_ids_index[recording.mbid] = lookup["spotify_track_ids"]
            for spotify_id in lookup["spotify_track_ids"]:
                spotify_id_mbid_index[spotify_id] = recording.mbid
    return recordings, mbid_spotify_ids_index, spotify_id_mbid_index


def _check_unplayable_tracks(sp: spotipy.Spotify, playlist_id: str):
    """ Retrieve tracks for given spotify playlist and split into lists of playable and unplayable tracks """
    playlist = sp.playlist_items(playlist_id, fields="items(track(name,id,is_playable))", market="from_token")
    playable = []
    unplayable = []
    for idx, item in enumerate(playlist["items"]):
        if item["track"]["is_playable"]:
            playable.append((idx, item["track"]["id"]))
        else:
            unplayable.append((idx, item["track"]["id"]))
    return playable, unplayable


def _get_alternative_track_ids(unplayable, mbid_spotify_id_idx, spotify_id_mbid_idx):
    """ For the list of unplayable track ids, find alternative tracks ids.

    mbid_spotify_id_idx is an index with mbid as key and list of equivalent spotify ids as value.
    spotify_id_mbid_idx is an index with spotify_id as key and the corresponding mbid as value.
    """
    index = defaultdict(list)
    spotify_ids = []
    
    for idx, spotify_id in unplayable:
        mbid = spotify_id_mbid_idx[spotify_id]
        other_spotify_ids = mbid_spotify_id_idx[mbid]

        for new_idx, new_spotify_id in enumerate(other_spotify_ids):
            if new_spotify_id == spotify_id:
                continue
            spotify_ids.append(new_spotify_id)
            index[idx].append(new_spotify_id)
            
    return spotify_ids, index


def _get_fixed_up_tracks(sp: spotipy.Spotify, spotify_ids, index):
    """ Lookup the all alternative spotify track ids, filter playable ones and if multiple track ids
        for same item match, prefer the one occurring earlier. If no alternative is playable, ignore the
        item altogether.
    """
    new_tracks = sp.tracks(spotify_ids, market="from_token")
    new_tracks_ids = set()
    for item in new_tracks["tracks"]:
        if item["is_playable"]:
            new_tracks_ids.add(item["id"])

    fixed_up_items = []
    for idx, spotify_ids in index.items():
        for spotify_id in spotify_ids:
            if spotify_id in new_tracks_ids:
                fixed_up_items.append((idx, spotify_id))
                break

    return fixed_up_items


def fixup_spotify_playlist(sp: spotipy.Spotify, playlist_id: str, mbid_spotify_id_idx, spotify_id_mbid_idx):
    """ Fix unplayable tracks in the given spotify playlist.

    Given a spotify playlist id, look it up and find unplayable tracks. If there are any unplayable tracks, try
    alternative spotify track ids from index/reverse_index if available. If alternative is not available or alternatives
    also are unplayable, remove the track entirely from the playlist. Finally, update the playlist if needed.
    """
    playable, unplayable = _check_unplayable_tracks(sp, playlist_id)
    if not unplayable:
        return

    alternative_ids, index = _get_alternative_track_ids(unplayable, mbid_spotify_id_idx, spotify_id_mbid_idx)
    if not alternative_ids:
        return

    fixed_up = _get_fixed_up_tracks(sp, alternative_ids, index)
    all_items = []
    all_items.extend(playable)
    all_items.extend(fixed_up)

    # sort all items by index value to ensure the order of tracks of original playlist is preserved
    all_items.sort(key=lambda x: x[0])
    # update all track ids the spotify playlist
    finalized_ids = [x[1] for x in all_items]

    # clear existing playlist
    sp.playlist_replace_items(playlist_id, [])
    # spotify API allows a max of 100 tracks in 1 request
    for chunk in chunked(finalized_ids, 100):
        sp.playlist_add_items(playlist_id, chunk)


def submit_to_spotify(spotify, playlist, spotify_user_id: str, is_public: bool = True,
                      is_collaborative: bool = False, existing_url: str = None):
    """ Submit or update an existing spotify playlist.

    If existing urls are specified then is_public and is_collaborative arguments are ignored.
    """
    filtered_recordings = [r for r in playlist.recordings if r.mbid]
    _, mbid_spotify_index, spotify_mbid_index = lookup_spotify_ids(filtered_recordings)
    spotify_track_ids = [r.spotify_id for r in filtered_recordings if r.spotify_id]
    if len(spotify_track_ids) == 0:
        return None, None

    logger.info("submit %d tracks" % len(spotify_track_ids))

    # Truncate to character limit for title
    playlist_name = playlist.name[0:100]
    # Remove HTML tags
    playlist_description = re.sub(
        CLEAN_HTML_RE, '', playlist.description or "")
    # Remove newlines
    playlist_description = ''.join(playlist_description.splitlines())
    # Truncate to character limit for description
    playlist_description = playlist_description[0:300]

    playlist_id, playlist_url = None, None
    if existing_url:
        # update existing playlist
        playlist_url = existing_url
        playlist_id = existing_url.split("/")[-1]
        try:
            spotify.playlist_change_details(
                playlist_id=playlist_id, name=playlist_name, description=playlist_description)
        except SpotifyException as err:
            # one possibility is that the user has deleted the spotify from playlist, so try creating a new one
            logger.info("provided playlist url has been unfollowed/deleted by the user, creating a new one")
            playlist_id, playlist_url = None, None

    if not playlist_id:
        # create new playlist
        spotify_playlist = spotify.user_playlist_create(
            user=spotify_user_id,
            name=playlist_name,
            public=is_public,
            collaborative=is_collaborative,
            description=playlist_description
        )
        playlist_id = spotify_playlist["id"]
        playlist_url = spotify_playlist["external_urls"]["spotify"]
    else:
        # existing playlist, clear it
        spotify.playlist_replace_items(playlist_id, [])

    # spotify API allows a max of 100 tracks in 1 request
    for chunk in chunked(spotify_track_ids, 100):
        spotify.playlist_add_items(playlist_id, chunk)

    fixup_spotify_playlist(spotify, playlist_id, mbid_spotify_index, spotify_mbid_index)

    playlist.add_metadata({"external_urls": {"spotify": playlist_url}})

    return playlist_url, playlist_id


def get_tracks_from_spotify_playlist(spotify_token, playlist_id):
    """ Get tracks from the Spotify playlist.
    """
    sp = spotipy.Spotify(auth=spotify_token, requests_timeout=10, retries=10)
    playlist_info = sp.playlist(playlist_id)

    offset = 0
    tracks = []

    # spotipy limits to 100 items for each call, so run iteratively with an offset until there are no more tracks
    while True:
        results = sp.playlist_items(playlist_id, limit=100, offset=offset)
        if len(results['items']) == 0:
            break

        tracks.extend(results['items'])
        # set new offset for the next loop
        offset = offset + len(results['items'])

    name = playlist_info["name"]
    description = playlist_info["description"]
    
    tracks = _convert_spotify_tracks_to_json(tracks)
    return tracks, name, description


def _convert_spotify_tracks_to_json(spotify_tracks: list):
    tracks = []
    for track in spotify_tracks:
        artists = track["track"].get("artists", [])
        artist_names = []
        for a in artists:
            name = a.get("name")
            if name is not None:
                artist_names.append(name)
        artist_name = ", ".join(artist_names)
        tracks.append({
            "recording_name": track["track"]["name"],
            "artist_name": artist_name,
        })
    return tracks



from libsonic import Connection


class FixedConnection(Connection):
    """
        This hack enables album ids to strings -- a PR has been submitted, but in case
        it doesn't get accepted in a timely manner, this workaround allows us to 
        continue.
    """

    def getAlbumInfo2(self, aid):
        """
        since 1.14.0

        Same as getAlbumInfo, but uses ID3 tags

        aid:int     The album ID    
        """
        methodName = 'getAlbumInfo2'
        viewName = '%s.view' % methodName

        # The release version has an int() cast here
        q = {'id': aid}
        req = self._getRequest(viewName, q)
        res = self._doInfoReq(req)
        self._checkStatus(res)
        return res



import mutagen.mp4

from troi.content_resolver.formats.tag_utils import get_tag_value, extract_track_number


EXTENSIONS = {'.m4a', '.m4b', '.m4p', '.m4v', '.m4r', '.mp4'}
READER = mutagen.mp4.MP4


def get_metadata(tags):

    mdata = {}
    mdata["artist_name"] = get_tag_value(tags, "©ART")
    mdata["artist_sortname"] = get_tag_value(tags, "soar", mdata["artist_name"])
    mdata["release_name"] = get_tag_value(tags, "©alb")
    mdata["recording_name"] = get_tag_value(tags, "©nam")
    mdata["track_num"] = extract_track_number(get_tag_value(tags, "trkn"))
    mdata["disc_num"] = get_and_decode(tags, "----:com.apple.iTunes:DISCSUBTITLE")
    mdata["artist_mbid"] = get_and_decode(tags, "----:com.apple.iTunes:MusicBrainz Artist Id")
    mdata["recording_mbid"] = get_and_decode(tags, "----:com.apple.iTunes:MusicBrainz Track Id")
    mdata["release_mbid"] = get_and_decode(tags, "----:com.apple.iTunes:MusicBrainz Album Id")
    mdata["duration"] = int(tags.info.length * 1000)

    return mdata


def get_and_decode(tags, tag_name):
    tag_value = get_tag_value(tags, tag_name)
    if tag_value is not None:
        tag_value = tag_value.decode("utf-8")
    return tag_value



from collections import defaultdict
from datetime import datetime
from random import randint


class plist(list):
    """ 
        This class implements a list object with percent based indexing and slicing:

        > a = plist((1,2,3,4))
        > a[50:]
        [3, 4]
        > a[25:75]
        [2, 3]
       
        Also supported are slicing on unit interval, using the uslice function:
        > a[a.uslice(.5)
        [3, 4]

    """

    def _get_index(self, percent):
        if len(self) == 0:
            return None

        if percent is not None:
            if isinstance(percent, int):
                if percent < 0 or percent > 100:
                    raise ValueError("Percent must be between 0 and 100.")
                return percent * len(self) // 100
            elif isinstance(percent, float):
                return int(percent * len(self))
            else:
                raise ValueError("percent must be an int or float.")
        else:
            return None

    def __getitem__(self, percent):

        if isinstance(percent, slice):
            start = self._get_index(percent.start)
            stop = self._get_index(percent.stop)
            return super().__getitem__(slice(self._get_index(percent.start), self._get_index(percent.stop)))
        else:
            return super().__getitem__(self._get_index(percent))

    def uslice(self, start_percent=None, stop_percent=None):

        if start_percent is None and stop_percent is None:
            raise IndexError("Invalid slice arguments. Must be at least one integer or float.")

        start_index = self._get_index(start_percent)
        stop_index = self._get_index(stop_percent)

        return super().__getitem__(slice(start_index, stop_index))

    def dslice(self, start=None, stop=None):
        return super().__getitem__(slice(start, stop))

    def random_item(self, start_percent=0, stop_percent=99, count=1):
        """
            Return a random item from the specified percent slice

            count specifies the number of items to return, default 1.

            Return value is a data item, unless count is > 1, then a list is returned
        """

        if len(self) == 0:
            return []

        start_index = self._get_index(start_percent)
        stop_index = self._get_index(stop_percent)

        data = super().__getitem__(slice(start_index, stop_index))
        items = []
        for i in range(len(data)):
            index = randint(0, len(data) - 1)
            items.append(data[index])
            del data[index]
            if len(items) == count or not data:
                break

        if count > 1:
            return items
        else:
            return items[0] if len(items) > 0 else []



from random import shuffle
from time import sleep

import requests

import troi
from troi import Recording
from troi import TARGET_NUMBER_OF_RECORDINGS


class LBRadioCollectionRecordingElement(troi.Element):
    """
        Given an MB recording collection, fetch it and randomly include recordings from it. mode does not
        apply to this element.
    """

    NUM_RECORDINGS_TO_COLLECT = TARGET_NUMBER_OF_RECORDINGS * 2

    def __init__(self, mbid, mode="easy"):
        troi.Element.__init__(self)
        self.mbid = mbid
        self.mode = mode

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def read(self, entities):

        # Fetch collection recordings
        params = {"collection": self.mbid, "fmt": "json"}

        while True:
            r = requests.get("https://musicbrainz.org/ws/2/recording", params=params)
            if r.status_code == 404:
                raise RuntimeError(f"Cannot find collection {self.mbid}.")

            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch collection {self.mbid}. {r.text}")

            break

        # Give feedback about what we collected
        self.local_storage["data_cache"]["element-descriptions"].append(f"collection {self.mbid}")

        # Fetch the recordings, then shuffle
        mbid_list = []
        for r in r.json()["recordings"]:
            if not r["video"]:
                mbid_list.append(r["id"])
        shuffle(mbid_list)

        # Select and convert the first n MBIDs into Recordings
        recordings = []
        for mbid in mbid_list[:self.NUM_RECORDINGS_TO_COLLECT]:
            recordings.append(Recording(mbid=mbid))

        return recordings



from sys import maxsize
from operator import attrgetter

import troi
from troi import PipelineError, Recording


class YearSortElement(troi.Element):
    '''
        Sort recordings by year ascending -- recordings that have the same
        year will be returned in an undefined order. Recordings
        that have no year set will be returned at the end of the list.
        If reverse=True, sort descending and return tracks with no year
        first.

        :param reverse: Reverse the sort order.
    '''

    def __init__(self, reverse=False):
        super().__init__()
        self.reverse = reverse

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        def year_sorter(r):
            if not r.year:
                return maxsize
            return r.year

        return sorted(inputs[0], key=year_sorter, reverse=self.reverse)



import json

from troi import Playlist
from troi.patch import Patch
from troi.playlist import RecordingsFromMusicServiceElement, PlaylistMakerElement
from troi.musicbrainz.recording_lookup import RecordingLookupElement
from troi.tools.apple_lookup import get_tracks_from_apple_playlist
from troi.tools.spotify_lookup import get_tracks_from_spotify_playlist
from troi.tools.soundcloud_lookup import get_soundcloud_playlist


class ImportPlaylistPatch(Patch):

    @staticmethod
    def inputs():
        """
        A patch that retrieves an existing playlist from Spotify for use in Troi.

        \b
        MS_TOKEN is the music service token from which the playlist is retrieved. For now, only Spotify tokens are accepted. 
        PLAYLIST_ID is the playlist id to retrieve the tracks from it.
        MUSIC_SERVICE is the music service from which the playlist is retrieved
        APPLE_USER_TOKEN is the apple user token. Optional, if music service is not Apple Music
        """
        return [
            {"type": "argument", "args": ["ms_token"], "kwargs": {"required": False}},
            {"type": "argument", "args": ["playlist_id"], "kwargs": {"required": False}},
            {"type": "argument", "args": ["music_service"], "kwargs": {"required": False}},
            {"type": "argument", "args": ["apple_user_token"], "kwargs": {"required": False}},
        ]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "import-playlist"

    @staticmethod
    def description():
        return "Retrieve a playlist from the Music Services (Spotify/Apple Music/SoundCloud)"

    def create(self, inputs):

        ms_token = inputs["ms_token"]
        playlist_id = inputs["playlist_id"]
        music_service = inputs["music_service"]
        apple_user_token = inputs["apple_user_token"]

        if apple_user_token == "":
            apple_user_token = None

        if music_service == "apple_music" and apple_user_token is None:
            raise RuntimeError("Authentication is required")

        playlist_fetchers = {
            "spotify": get_tracks_from_spotify_playlist,
            "apple_music": get_tracks_from_apple_playlist,
            "soundcloud": get_soundcloud_playlist
        }

        get_tracks = playlist_fetchers.get(music_service)
        if not get_tracks:
            raise ValueError(f"Unsupported music service: {music_service}")

        _, name, desc = get_tracks(ms_token, apple_user_token, playlist_id) if music_service == "apple_music" else get_tracks(ms_token, playlist_id)

        source = RecordingsFromMusicServiceElement(token=ms_token, playlist_id=playlist_id, music_service=music_service, apple_user_token=apple_user_token)
        
        rec_lookup = RecordingLookupElement()
        rec_lookup.set_sources(source)

        pl_maker = PlaylistMakerElement(name, desc, patch_slug=self.slug())
        pl_maker.set_sources(rec_lookup)

        return pl_maker



import logging

import requests
from more_itertools import chunked

from troi.tools.apple_lookup import get_tracks_from_apple_playlist
from troi.tools.spotify_lookup import get_tracks_from_spotify_playlist
from troi.tools.soundcloud_lookup import get_tracks_from_soundcloud_playlist

MAX_LOOKUPS_PER_POST = 50
MBID_LOOKUP_URL = "https://api.listenbrainz.org/1/metadata/lookup/"

logger = logging.getLogger(__name__)


def music_service_tracks_to_mbid(token, playlist_id, music_service, apple_user_token=None):
    """ Convert Spotify playlist tracks to a list of MBID tracks.
    """
    if music_service == "spotify":
        tracks, name, desc = get_tracks_from_spotify_playlist(token, playlist_id)
    elif music_service == "apple_music":
        tracks, name, desc = get_tracks_from_apple_playlist(token, apple_user_token, playlist_id)
    elif music_service == "soundcloud":
        tracks = get_tracks_from_soundcloud_playlist(token, playlist_id)
    else:
        raise ValueError("Unknown music service")

    track_lists = list(chunked(tracks, MAX_LOOKUPS_PER_POST))
    return mbid_mapping_tracks(track_lists)


def mbid_mapping_tracks(track_lists):
    """ Given a track_name and artist_name, try to find MBID for these tracks from mbid lookup.
    """
    track_mbids = []
    for tracks in track_lists:
        params = {
            "recordings": tracks
        }
        response = requests.post(MBID_LOOKUP_URL, json=params)
        if response.status_code == 200:
            data = response.json()
            for d in data:
                if d is not None and "recording_mbid" in d:
                    track_mbids.append(d["recording_mbid"])
        else:
            logger.error("Error occurred: %s", response.text)
    return track_mbids
